#include "SADB.h"
using namespace SADB;

/// Constructor, initializes object
/// @see Blackboard::Initialize
Blackboard::Blackboard(void){Initialize();}

/// Destructor, deinitializes object
/// @see Blackboard::Free
Blackboard::~Blackboard(void){Free();}

/// Initializes object
void Blackboard::Initialize(void){
 Address=NULL;
}

/// Frees allocated memory
void Blackboard::Free(void){
 if(Address!=NULL)free(Address);Address=NULL;
}

/// Connect to the SADB Server
int Blackboard::ConnectToServer(void){
  return ConnectToServer(SADBAddress,SADBPort);
}

/// Connect to the SADB Server
/// @param Address
///  Address of the server
/// @param Port
///  Port of the server
int Blackboard::ConnectToServer(const char * address,int port){
 Port=port;
 Address=(char*)realloc(Address,(strlen(address)+1)*sizeof(char));
 if(Address==NULL)Error("Blackboard","Cannot allocate memory to store address");
 memcpy(Address,address,strlen(address)+1);
 return SADB.ConnectTo(address,port);
}


/// Connect to the SADB Server and retry to connect until the connection is established
void Blackboard::ConnectToServerWithRetry(void){
 ConnectToServerWithRetry(SADBAddress,SADBPort);
}

/// Connect to the SADB Server and retry to connect until the connection is established 
/// @param Address
///  Address of the server
/// @param Port
///  Port of the server
void Blackboard::ConnectToServerWithRetry(const char * address,int port){
 Port=port;
 Address=(char*)realloc(Address,(strlen(address)+1)*sizeof(char));
 if(Address==NULL)Error("Blackboard","Cannot allocate memory to store address");
 memcpy(Address,address,strlen(address)+1);
 int c=1;
 while(c)c=SADB.ConnectTo(address,port);
}


/// Disconnect from the SADB Server
void Blackboard::Disconnect(void){
 SADB.Close();
}



/// Ping the SADB Server for testing purposes and to determine latency
/// @return number of milliseconds between the ping and the pong 
SADBs64 Blackboard::Ping(void){
 Timestamp Ping,Pong;Ping.Now();
// Ping.Print();printf("\n");
 SADBu8 * Buffer=(SADBu8*)malloc(SADBHeaderLength);
 GenerateSADBHeader(Buffer,0x01,0,0,Ping.Second,Ping.Millisecond,0);
 SADB.Send(Buffer,SADBHeaderLength);
 SADB.ReceiveExactly(Buffer,SADBHeaderLength);
 Pong.Second=*((SADBu64*)(Buffer+0x13));
 Pong.Millisecond=*((SADBu16*)(Buffer+0x1B));
 free(Buffer);
// Pong.Print();printf("\n");
 return Ping.MillisecondDifference(&Pong);
}

/// Find IDs of objects matching a query
/// @note Make sure the pointer to the list is initialized to NULL
/// @param Request
///  String representation of the request
/// @param Objects
///  Pointer to a list of objet IDs
/// @return Number of Objects matching the query
int Blackboard::FindObjects(const char *Request,SADBu64**Objects){
 SendSADBPacket(SADB.Socket,0x17,0,0,0,0,strlen(Request)+1,(SADBu8*)Request);//+1 to include the \0

 SADBu8 Command=0;
 SADBu64 OUID,Argument;
 Timestamp Time;
 SADBu32 DataLength;
 SADBu8 * Data=NULL;

 SADBReceive(SADB.Socket,&Command,&OUID,&Argument,&(Time.Second),&(Time.Millisecond),&DataLength,&Data);

 if(OUID!=0){
  *Objects=(SADBu64*)realloc(*Objects,sizeof(SADBu64));//single match
  (*Objects)[0]=OUID;
  free(Data);
  return 1;
 }else if(DataLength!=0){//multiple matches
  *Objects=(SADBu64*)realloc(*Objects,DataLength);
  memcpy(*Objects,Data,DataLength);
  return DataLength/sizeof(SADBu64);
 }else{//no matches
  return 0;
 }


}

/// Check if an Object exists (has been created and not deleted)
/// @param OUID
///  ID of the Object
/// @return 1 if CUID is valid, 0 otherwise
int Blackboard::IsValidObject(SADBu64 OUID){
 SendSADBPacket(SADB.Socket,0x20,OUID,0,0,0,0,NULL);
 SADBu8 CMD;SADBu64 ID;SADBu64 Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data);
 if(Data!=NULL)free(Data);//Just in case
 return (int)Argument;
}


/// Create a new category on the blackboard
/// @return CUID of the new category
SADBu64 Blackboard::CreateCategory(void){
 SADBu64 ID;
 SendSADBPacket(SADB.Socket,0x0E,0,0,0,0,0,0);
 SADBReceiveID(SADB.Socket,&ID);
 return ID;
}

/// Get the ID of a category given its name.  If the category does not exist, create a new one and return its ID.
/// @param Name
///  Name of the category
/// @return CUID of the (possibly new) category
SADBu64 Blackboard::CreateCategoryIfNonExistant(const char * Name){
 SADBu64 CUID=GetCategoryByName(Name);
 if(CUID==0){
  CUID=CreateCategory();
  SetCategoryName(CUID,Name);
 }
 return CUID;
}

/// Delete an existing category from the blackboard
/// @param CUID
///  ID of the category to remove
void Blackboard::DeleteCategory(SADBu64 CUID){
 SendSADBPacket(SADB.Socket,0x0F,CUID,0,0,0,0,0);
 SADBWaitForResponse(SADB.Socket);
}

/// Check if a Category exists (has been created and not deleted)
/// @param CUID
///  ID of the category
/// @return 1 if CUID is valid, 0 otherwise
int Blackboard::IsValidCategory(SADBu64 CUID){
 SendSADBPacket(SADB.Socket,0x21,CUID,0,0,0,0,NULL);
 SADBu8 CMD;SADBu64 ID;SADBu64 Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data);
 if(Data!=NULL)free(Data);//Just in case
 return (int)Argument;
}


/// Set the name of a category on the blackboard
/// @param CUID
///  ID of the category to edit
/// @param Name
///  Null-terminated string to set
void Blackboard::SetCategoryName(SADBu64 CUID,const char * Name){
 SendSADBPacket(SADB.Socket,0x10,CUID,0,0,0,strlen(Name)+1,(SADBu8*)Name);//+1 to include the \0
 SADBWaitForResponse(SADB.Socket);
}

/// Get the name of a category on the blackboard
/// @param CUID
///  ID of the target category
/// @param Name
///  Return address
void Blackboard::GetCategoryName(SADBu64 CUID,SADBu8 ** Name){
 SendSADBPacket(SADB.Socket,0x11,CUID,0,0,0,0,NULL);
 SADBReceiveString(SADB.Socket,Name);
}

/// Set the description of a category on the blackboard
/// @param CUID
///  ID of the category to edit
/// @param Description
///  Null-terminated string to set
void Blackboard::SetCategoryDescription(SADBu64 CUID,const char * Description){
 SendSADBPacket(SADB.Socket,0x12,CUID,0,0,0,strlen(Description)+1,(SADBu8*)Description);
 SADBWaitForResponse(SADB.Socket);
}

/// Get the Description of a category on the blackboard
/// @param CUID
///  ID of the target category
/// @param Name
///  Return address
void Blackboard::GetCategoryDescription(SADBu64 CUID,SADBu8 ** Description){
 SendSADBPacket(SADB.Socket,0x13,CUID,0,0,0,0,NULL);
 SADBReceiveString(SADB.Socket,Description);
}

/// Get the CUID of a category given its name
/// @param Name
///  Name of the category
/// @output CUID of the object, or 0 if not found
SADBu64 Blackboard::GetCategoryByName(const char * Name){
 SADBu64 ID;
 SendSADBPacket(SADB.Socket,0x14,0,0,0,0,strlen(Name)+1,(SADBu8*)Name);//+1 for the '\0'
 SADBReceiveID(SADB.Socket,&ID);
 return ID;
}

/// Add an object to a category
/// @param CUID
///  ID of the category
/// @param OUID
///  ID of the object
void Blackboard::AddObjectToCategory(SADBu64 CUID,SADBu64 OUID){
 SendSADBPacket(SADB.Socket,0x15,CUID,OUID,0,0,0,0);
 SADBWaitForResponse(SADB.Socket);
}

/// Remove an object from a category
/// @param CUID
///  ID of the category
/// @param OUID
///  ID of the object
void Blackboard::RemoveObjectFromCategory(SADBu64 CUID,SADBu64 OUID){
 SendSADBPacket(SADB.Socket,0x16,CUID,OUID,0,0,0,0);
 SADBWaitForResponse(SADB.Socket);
}


/// Create a new object on the blackboard
/// @return OUID of the new object
SADBu64 Blackboard::CreateObject(void){
 SADBu64 ID;
 SendSADBPacket(SADB.Socket,0x02,0,0,0,0,0,0);
 SADBReceiveID(SADB.Socket,&ID);
 return ID;
}

/// delete an existing object from the blackboard
/// @param OUID
///  ID of the object to remove
void Blackboard::DeleteObject(SADBu64 OUID){
 SendSADBPacket(SADB.Socket,0x03,OUID,0,0,0,0,0);
 SADBWaitForResponse(SADB.Socket);
}

/// Set the name of an object on the blackboard
/// @param OUID
///  ID of the object to edit
/// @param Name
///  Null-terminated string to set
void Blackboard::SetObjectName(SADBu64 OUID,const char * Name){
 SendSADBPacket(SADB.Socket,0x04,OUID,0,0,0,strlen(Name)+1,(SADBu8*)Name);//+1 to include the \0
 SADBWaitForResponse(SADB.Socket);
}

/// Get the name of an object on the blackboard
/// @param OUID
///  ID of the target object
/// @param Name
///  Return address
void Blackboard::GetObjectName(SADBu64 OUID,SADBu8 ** Name){
 SendSADBPacket(SADB.Socket,0x05,OUID,0,0,0,0,NULL);
 SADBReceiveString(SADB.Socket,Name);
}


/// Set the description of an object on the blackboard
/// @param OUID
///  ID of the object to edit
/// @param Description
///  Null-terminated string to set
void Blackboard::SetObjectDescription(SADBu64 OUID,const char * Description){
 SendSADBPacket(SADB.Socket,0x06,OUID,0,0,0,strlen(Description)+1,(SADBu8*)Description);
 SADBWaitForResponse(SADB.Socket);
}

/// Get the Description of an object on the blackboard
/// @param OUID
///  ID of the target object
/// @param Name
///  Return address
void Blackboard::GetObjectDescription(SADBu64 OUID,SADBu8 ** Description){
 SendSADBPacket(SADB.Socket,0x07,OUID,0,0,0,0,NULL);
 SADBReceiveString(SADB.Socket,Description);
}

/// Get the OUID of an object given its name
/// @param Name
///  Name of the object
/// @output OUID of the object, or 0 if not found
SADBu64 Blackboard::GetObjectByName(const char * Name){
 SADBu64 ID;
 SendSADBPacket(SADB.Socket,0x08,0,0,0,0,strlen(Name)+1,(SADBu8*)Name);//+1 for the '\0'
 SADBReceiveID(SADB.Socket,&ID);
 return ID;
}

/// Reset the SADB Server
/// @note Should probably never be used
void Blackboard::Reset(void){
 SendSADBPacket(SADB.Socket,0x0D,0,0,0,0,0,NULL);
 SADBWaitForResponse(SADB.Socket);
}

/// Send the value of an object
/// @param OUID
///  ID of the object in SADB
/// @param Object
///  Pointer to an object
/// @return Returns 0 on success
int Blackboard::SendObjectValue(SADBu64 OUID, BlackboardObject * Object){
 int r=1;
 if(OUID!=0){//If non-zero ID
  //Send Packet, Argument is the data type, the first 4 bytes of data is the number of dimensions
  //the next n*4 bytes are the dimensions, then the data
//  if(Object->Dimensionality>0){//if the object contains data
   SADBu32 * Buffer=(SADBu32*)malloc((Object->Dimensionality+1)*sizeof(SADBu32));
   Buffer[0]=(SADBu32)Object->Dimensionality;
   for(SADBu32 i=0;i<Object->Dimensionality;i++)
    Buffer[i+1]=(SADBu32)Object->Dimensions[i];
   r=SendSADBPacket(SADB.Socket,0x09,OUID,DataTypeEnumToInt(Object->DataType),(Object->Time).Second,(Object->Time).Millisecond,
    (Object->Dimensionality+1)*sizeof(SADBu32),(SADBu8 *)Buffer,//Dimension information
    Object->TotalNumberOfElements*Object->ElementSize,(SADBu8*)Object->Data);//Actual Data
   free(Buffer);
   //Wait for response
   SADBu8 CMD;SADBu64 ID,Argument;
   SADBu32 DataLength;SADBu8 *Data=NULL;
   SADBu64 Second;SADBu16 Millisecond;
   SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data);
   if(Data!=NULL)free(Data);//Just in case
//  }
 }
 return r;
}

/// Get the latest value of an object
/// @param OUID
///  ID of the object in SADB
/// @param Object
///  Pointer to an object that will contain the latest value
/// @return 0 on success, 1 otherwise
int Blackboard::GetLatestObjectValue(SADBu64 OUID, BlackboardObject * Object){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x0A,OUID,0,0,0,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 //Load data into object
 if(ID!=OUID){if(Data!=NULL)free(Data);return 1;}//Returned ID not correct
 SADBu32 D=((SADBu32*)Data)[0];
 Object->Time.Second=Second;
 Object->Time.Millisecond=Millisecond;
 Object->SetDataType(DataTypeIntToEnum((SADBu8)Argument));
 if(Object->SetDimensions(D,((SADBu32*)Data)+1)){if(Data!=NULL)free(Data);return 1;}//Could not set dimensions, free received data, return 1
 memcpy(Object->Data,Data+(D+1)*sizeof(SADBu32),Object->TotalNumberOfElements*Object->ElementSize);
 if(Data!=NULL)free(Data);//Free memory
 return 0;
}

/// Get the latest timestamp of an object
/// @note This method does not get the value of the object, only the timestamp
/// @param OUID ID of the object in SADB
/// @param time Pointer to a Timestamp object
/// @return 0 on success, 1 otherwise
int Blackboard::GetLatestObjectTimestamp(SADBu64 OUID, Timestamp * time){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x1A,OUID,0,0,0,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 if(ID!=0){
  time->Second=Second;
  time->Millisecond=Millisecond;
 }else{if(Data!=NULL)free(Data);return 1;}//Returned ID not valid
 if(Data!=NULL)free(Data);//just in case
 return 0;
}


/// Get the next timestamp of an object
/// @note This method does not get the value of the object, only the timestamp
/// @param OUID ID of the object in SADB
/// @param Current Pointer to a Timestamp object
/// @param[out] Next Pointer to a Timestamp object to store result
/// @return 0 on success, 1 otherwise
int Blackboard::GetNextObjectTimestamp(SADBu64 OUID,Timestamp*Current,Timestamp*Next){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x22,OUID,0,Current->Second,Current->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 if(ID!=0){
  Next->Second=Second;
  Next->Millisecond=Millisecond;
 }else{if(Data!=NULL)free(Data);return 1;}//Returned ID not valid
 if(Data!=NULL)free(Data);//just in case
 return 0;
}

/// Get the previous timestamp of an object
/// @note This method does not get the value of the object, only the timestamp
/// @param OUID ID of the object in SADB
/// @param Current Pointer to a Timestamp object
/// @param[out] Previous Pointer to a Timestamp object to store result
/// @return 0 on success, 1 otherwise
int Blackboard::GetPreviousObjectTimestamp(SADBu64 OUID,Timestamp*Current,Timestamp*Previous){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x23,OUID,0,Current->Second,Current->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 if(ID!=0){
  Previous->Second=Second;
  Previous->Millisecond=Millisecond;
 }else{if(Data!=NULL)free(Data);return 1;}//Returned ID not valid
 if(Data!=NULL)free(Data);//just in case
 return 0;
}

/// Get the nearest timestamp of an object to the target
/// @note This method does not get the value of the object, only the timestamp
/// @param OUID ID of the object in SADB
/// @param Target Pointer to a Timestamp object
/// @param[out] Result Pointer to a Timestamp object to store result
/// @return 0 on success, 1 otherwise
int Blackboard::GetNearestObjectTimestamp(SADBu64 OUID,Timestamp*Target,Timestamp*Result){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x24,OUID,0,Target->Second,Target->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 if(ID!=0){
  Result->Second=Second;
  Result->Millisecond=Millisecond;
 }else{if(Data!=NULL)free(Data);return 1;}//Returned ID not valid
 if(Data!=NULL)free(Data);//just in case
 return 0;
}


/// Get the latest value of an object at a given timestamp
/// @param OUID
///  ID of the object in SADB
/// @param Object
///  Pointer to an object that will contain the value
/// @param Time
///  Requested timestamp
/// @param Interpolation
///  Requested interpolation method
/// @param nPoints
///  Number of points to use for interpolation
/// @return 0 on success, 1 otherwise
int Blackboard::GetObjectValueAtTimestamp(SADBu64 OUID, BlackboardObject * Object,Timestamp * Time,int Interpolation,int nPoints){
 if(OUID==0)return 1;//Not a valid ID
 SADBu64 Argument=((nPoints&0xFF )<<8)|(Interpolation&0xFF);
 if(SendSADBPacket(SADB.Socket,0x0C,OUID,Argument,Time->Second,Time->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 //Load data into object
 if(ID!=OUID){if(Data!=NULL)free(Data);return 1;}//Returned ID not correct
 SADBu32 D=((SADBu32*)Data)[0];
 Object->Time.Second=Second;
 Object->Time.Millisecond=Millisecond;
 Object->SetDataType(DataTypeIntToEnum((SADBu8)Argument));
 if(Object->SetDimensions(D,((SADBu32*)Data)+1)){if(Data!=NULL)free(Data);return 1;}//Could not set dimensions, free received data, return 1
 memcpy(Object->Data,Data+(D+1)*sizeof(SADBu32),Object->TotalNumberOfElements*Object->ElementSize);
 if(Data!=NULL)free(Data);//Free memory
 return 0;
}


/// Get the value of an object at the nearest Timestamp
/// @param OUID
///  ID of the object in SADB
/// @param Time
///  Requested Timestamp
/// @param Object
///  Pointer to an object that will contain the latest value
/// @return 0 on success, 1 otherwise
int Blackboard::GetObjectValueAtNearestTimestamp(SADBu64 OUID,Timestamp * Time, BlackboardObject * Object){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x0B,OUID,0,Time->Second,Time->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 //Load data into object
 if(ID!=OUID){if(Data!=NULL)free(Data);return 1;}//Returned ID not correct
 SADBu32 D=((SADBu32*)Data)[0];
 Object->Time.Second=Second;
 Object->Time.Millisecond=Millisecond;
 Object->SetDataType(DataTypeIntToEnum((SADBu8)Argument));
 if(Object->SetDimensions(D,((SADBu32*)Data)+1)){if(Data!=NULL)free(Data);return 1;}//Could not set dimensions, free received data, return 1
 memcpy(Object->Data,Data+(D+1)*sizeof(SADBu32),Object->TotalNumberOfElements*Object->ElementSize);
 if(Data!=NULL)free(Data);//Free memory
 return 0;
}


/// Tell the SADB server to discard the oldest value of an object.
/// @param OUID
///  ID of the target object
/// @return 0 on success, 1 otherwise
int Blackboard::RemoveOldestValue(SADBu64 OUID){
 if(SendSADBPacket(SADB.Socket,0x18,OUID,0,0,0,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 if(SADBWaitForResponse(SADB.Socket)!=0)return 1;//Wait for response
 return 0;
}

/// Tell the SADB server to discard older values of an object and to keep only the nValues latest values.
/// @param OUID
///  ID of the target object
/// @param nValues
///  number of values to keep
/// @return 0 on success, 1 otherwise
int Blackboard::KeepOnlyLatestValues(SADBu64 OUID,int nValues){
 if(SendSADBPacket(SADB.Socket,0x19,OUID,nValues,0,0,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 if(SADBWaitForResponse(SADB.Socket)!=0)return 1;//Wait for response
 return 0;
}


/// Get the value of an object at the next Timestamp after the target time
/// @note if there is no timestamp after the target time, the last value (latest) is returned
/// @param OUID
///  ID of the object in SADB
/// @param Time
///  Target Timestamp
/// @param Object
///  Pointer to an object that will contain the value
/// @return 0 on success, 1 otherwise
int Blackboard::GetObjectValueAtNextTimestamp(SADBu64 OUID,Timestamp * Time, BlackboardObject * Object){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x1B,OUID,0,Time->Second,Time->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 //Load data into object
 if(ID!=OUID){if(Data!=NULL)free(Data);return 1;}//Returned ID not correct
 SADBu32 D=((SADBu32*)Data)[0];
 Object->Time.Second=Second;
 Object->Time.Millisecond=Millisecond;
 Object->SetDataType(DataTypeIntToEnum((SADBu8)Argument));
 if(Object->SetDimensions(D,((SADBu32*)Data)+1)){if(Data!=NULL)free(Data);return 1;}//Could not set dimensions, free received data, return 1
 memcpy(Object->Data,Data+(D+1)*sizeof(SADBu32),Object->TotalNumberOfElements*Object->ElementSize);
 //Free memory
 if(Data!=NULL)free(Data);
 return 0;
}


/// Get the value of an object at the next Timestamp after the target time
/// @note if there is no timestamp before the target time, the first value (earliest) is returned
/// @param OUID
///  ID of the object in SADB
/// @param Time
///  Target Timestamp
/// @param Object
///  Pointer to an object that will contain the value
/// @return 0 on success, 1 otherwise
int Blackboard::GetObjectValueAtPreviousTimestamp(SADBu64 OUID,Timestamp * Time, BlackboardObject * Object){
 if(OUID==0)return 1;//Not a valid ID
 if(SendSADBPacket(SADB.Socket,0x1C,OUID,0,Time->Second,Time->Millisecond,0,NULL)!=0)return 1;//Send Request, return 1 if it fails
 //Wait for response
 SADBu8 CMD;SADBu64 ID,Argument;
 SADBu32 DataLength;SADBu8 *Data=NULL;
 SADBu64 Second;SADBu16 Millisecond;
 if(SADBReceive(SADB.Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data)!=0){if(Data!=NULL)free(Data);return 1;}
 //Load data into object
 if(ID!=OUID){if(Data!=NULL)free(Data);return 1;}//Returned ID not correct
 SADBu32 D=((SADBu32*)Data)[0];
 Object->Time.Second=Second;
 Object->Time.Millisecond=Millisecond;
 Object->SetDataType(DataTypeIntToEnum((SADBu8)Argument));
 if(Object->SetDimensions(D,((SADBu32*)Data)+1)){if(Data!=NULL)free(Data);return 1;}//Could not set dimensions, free received data, return 1
 memcpy(Object->Data,Data+(D+1)*sizeof(SADBu32),Object->TotalNumberOfElements*Object->ElementSize);
 //Free memory
 if(Data!=NULL)free(Data);
 return 0;
}


/// Get the next Object ID.
/// @return Next available Object ID, or 0 on error (0 is the ID number reserved for no object)
SADBu64 Blackboard::GetNextObjectID(void){
 SADBu64 ID;
 if(SendSADBPacket(SADB.Socket,0x1D,0,0,0,0,0,NULL)!=0)return 0;//Send Request
 if(SADBReceiveID(SADB.Socket,&ID)!=0)return 0;
 return ID;
}

/// Get the next Category ID.
/// @return Next available Category ID, or 0 on error (0 is the ID number reserved for no category)
SADBu64 Blackboard::GetNextCategoryID(void){
 SADBu64 ID;
 if(SendSADBPacket(SADB.Socket,0x1E,0,0,0,0,0,NULL)!=0)return 0;//Send Request
 if(SADBReceiveID(SADB.Socket,&ID)!=0)return 0;
 return ID;
}
