#include "SADB.h"
using namespace SADB;

/// Constructor, initializes object
/// @see BlackboardObject::Initialize
BlackboardObject::BlackboardObject(void){Initialize();}

/// Destructor, deinitializes object
/// @see BlackboardObject::Free
BlackboardObject::~BlackboardObject(void){Free();}

/// Initializes object
void BlackboardObject::Initialize(void){
 Data=NULL;Dimensions=NULL;
 Name=NULL;Description=NULL;
 Board=NULL;
 Time.Second=0;Time.Millisecond=0;
 Dimensionality=0;
 SetDataType(BOByte);//default
}

/// Frees allocated memory
void BlackboardObject::Free(void){
 //printf("Freeing %ld - %s\n",ID,Name);
 if(Data!=NULL)free(Data);Data=NULL;
 if(Dimensions!=NULL)free(Dimensions);Dimensions=NULL;
 if(Name!=NULL)free(Name);Name=NULL;
 if(Description!=NULL)free(Description);Description=NULL;
}

/// Associate the object to a blackboard
/// @note an object may only be associated with a single blackboard
/// @param board
///  Pointer to the target blackboard object
void BlackboardObject::SetBlackboard(Blackboard * board){
 Board=board;
}

/// Get the object Id from the associated blackboard from the name
/// @param name
///  Name of the object on the blackboard
void BlackboardObject::GetByName(const char * name){
 if(Board!=NULL)ID=Board->GetObjectByName(name);
}

/// Add the object to a blackboard
/// @note an object may only be associated with a single blackboard
/// @param board
///  Pointer to the target blackboard object
void BlackboardObject::AddToBlackboard(void){
 if(Board!=NULL){
  ID=Board->CreateObject();
  if(Name!=NULL)Board->SetObjectName(ID,(const char *)Name);
  if(Description!=NULL)Board->SetObjectDescription(ID,(const char *)Description);
 }
}

/// Send the object data to the blackboard
/// @return Returns 0 on success, will return 1 if either the board is not defined or the transfer fails
int BlackboardObject::Send(void){
 int r=1;
 if(Board!=NULL)r=Board->SendObjectValue(ID,this);
 return r;
}

/// Get the latest value form the SADB server
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetLatestValue(void){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 return Board->GetLatestObjectValue(ID,this);
}

/// Check to see if there is a new value on the SADB server, without retrieving it
/// @return 1 if a new value is present, 0 if no new value is available, and 2 on error
int BlackboardObject::CheckForNewValue(void){
 if(ID==0)return 2;//Invalid ID
 if(Board==NULL)return 2;//Not associated to a board
 Timestamp Latest;Latest.Second=0;Latest.Millisecond=0;//Create a timestamp object and clear it
 if(Board->GetLatestObjectTimestamp(ID,&Latest)!=0)return 2;
 if((Latest.Second!=0)&&(Latest.Millisecond!=0)){
  if(Time.MillisecondDifference(&Latest)>0)return 1;
 }
 return 0;
}

///Get the latest timestamp on the server
///@note This function only gets the timestamp, not the acutal data
///@param[out] Time Pointer to a Timestamp object to store result
///@return 0 on success, 1 otherwise
int BlackboardObject::GetLatestTimestamp(Timestamp*Time){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 2;//Not associated to a board
 if(Board->GetLatestObjectTimestamp(ID,Time)!=0)return 3;
 return 0;
}
///Get the previous timestamp on the server
///@note This function only gets the timestamp, not the acutal data
///@param Current Pointer to a Timestamp object
///@param[out] Previous Pointer to a Timestamp object to store result
///@return 0 on success, 1 otherwise
int BlackboardObject::GetPreviousTimestamp(Timestamp * Current,Timestamp*Previous){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 2;//Not associated to a board
 if(Board->GetPreviousObjectTimestamp(ID,Current,Previous)!=0)return 3;
 if(Current->MillisecondDifference(Previous)>=0)return 4; 
 return 0;
}
///Get the next timestamp on the server
///@note This function only gets the timestamp, not the acutal data
///@param Current Pointer to a Timestamp object
///@param[out] Next Pointer to a Timestamp object to store result
///@return 0 on success, 1 otherwise
int BlackboardObject::GetNextTimestamp(Timestamp * Current,Timestamp*Next){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 2;//Not associated to a board
 if(Board->GetNextObjectTimestamp(ID,Current,Next)!=0)return 3;
 if(Current->MillisecondDifference(Next)<=0)return 4;
 return 0;
}
///Get the timestamp on server closest to the rquested one
///@note This function only gets the timestamp, not the acutal data
///@param Target Pointer to a Timestamp object
///@param[out] Result Pointer to a Timestamp object to store result
///@return 0 on success, 1 otherwise
int BlackboardObject::GetNearestTimestamp(Timestamp * Target,Timestamp*Result){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 2;//Not associated to a board
 if(Board->GetNearestObjectTimestamp(ID,Target,Result)!=0)return 3;
 return 0;
}


/// Get the value at the nearest timestamp
/// @param Time Requested Timestamp
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtNearestTimestamp(Timestamp * Time){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 return Board->GetObjectValueAtNearestTimestamp(ID,Time,this);
}

/// Get the value at a given timestamp, if such a value exists
/// @param Time Requested Timestamp
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtTimestampNoInterpolation(Timestamp * Time){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 int r=Board->GetObjectValueAtNearestTimestamp(ID,Time,this);
 if(r!=0)return 2;
 if(0!=Time->MillisecondDifference(&(this->Time)))return 3;
 return 0;
}

/// Get the value at the exact timestamp with default interpolation method (linear)
/// @param Time
///  Requested Timestamp
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtTimestamp(Timestamp * Time){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 return Board->GetObjectValueAtTimestamp(ID,this,Time,Interpolation_Linear,0);
}


/// Get the value at the nearest timestamp
/// @param Time
///  Requested Timestamp
/// @param Interpolation
///  Requested Interpolation method
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtTimestamp(Timestamp * Time,int Interpolation){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 return Board->GetObjectValueAtTimestamp(ID,this,Time,Interpolation,0);
}

/// Get the value at the nearest timestamp
/// @param Time
///  Requested Timestamp
/// @param Interpolation
///  Requested Interpolation method
/// @param nPoints
///  Number of points to use for interpolation
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtTimestamp(Timestamp * Time,int Interpolation,int nPoints){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 return Board->GetObjectValueAtTimestamp(ID,this,Time,Interpolation,nPoints);
}

/// Get the value at the next timestamp
/// @param Time
///  Target Timestamp
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtNextTimestamp(Timestamp * Time){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 int R=Board->GetObjectValueAtNextTimestamp(ID,Time,this);
 if(Time->MillisecondDifference(&(this->Time))<=0)return 2;
 return R;
}

/// Get the value at the previous timestamp
/// @param Time
///  Target Timestamp
/// @return 0 on success, 1 otherwise
int BlackboardObject::GetValueAtPreviousTimestamp(Timestamp * Time){
 if(ID==0)return 1;//Invalid ID
 if(Board==NULL)return 1;//Not associated to a board
 int R=Board->GetObjectValueAtPreviousTimestamp(ID,Time,this);
 if(Time->MillisecondDifference(&(this->Time))>=0)return 2;
 return R;
}


/// Add the object to a category
/// @param CUID
///  ID of the category
void BlackboardObject::AddToCategory(SADBu64 CUID){
 if(Board!=NULL)Board->AddObjectToCategory(CUID,ID);
}

/// Add the object to a category
/// @param Cat
///  Name of the category
void BlackboardObject::AddToCategory(const char * Cat){
 if(Board!=NULL){
  SADBu64 CUID=Board->GetCategoryByName(Cat);//The CUID
  AddToCategory(CUID);
 }
}

/// Add the object to a category
/// @param CUID
///  ID of the category
void BlackboardObject::RemoveFromCategory(SADBu64 CUID){
 if(Board!=NULL)Board->RemoveObjectFromCategory(CUID,ID);
}

/// Add the object to a category
/// @param Cat
///  Name of the category
void BlackboardObject::RemoveFromCategory(const char * Cat){
 if(Board!=NULL){
  SADBu64 CUID=Board->GetCategoryByName(Cat);//The CUID
  RemoveFromCategory(CUID);
 }
}




/// Create an object (convenience function)
/// @param board
///  Pointer to the target blackboard object
/// @param name
///  Name of the object
void BlackboardObject::Create(Blackboard * board,const char * name){
 SetBlackboard(board);
 CreateIfNonExistant(name);
}


/// Create an object (convenience function)
/// @param board
///  Pointer to the target blackboard object
/// @param name
///  Name of the object
/// @param description
///  Description of the object
void BlackboardObject::Create(Blackboard * board,const char * name,const char * description){
 SetBlackboard(board);
 CreateIfNonExistant(name);
 SetDescription(description);
}

/// Create an object (convenience function)
/// @param board
///  Pointer to the target blackboard object
/// @param name
///  Name of the object
/// @param type
///  Datatype of the object
void BlackboardObject::Create(Blackboard * board,const char * name,enum BODataType type){
 SetBlackboard(board);
 CreateIfNonExistant(name);
 SetDataType(type);
}


/// Create an object (convenience function)
/// @param board
///  Pointer to the target blackboard object
/// @param name
///  Name of the object
/// @param description
///  Description of the object
/// @param type
///  Datatype of the object
void BlackboardObject::Create(Blackboard * board,const char * name,const char * description,enum BODataType type){
 SetBlackboard(board);
 CreateIfNonExistant(name);
 SetDescription(description);
 SetDataType(type);
}






/// Sets the type of data that the object will contain
/// @param Type
///   Data Type
void BlackboardObject::SetDataType(enum BODataType Type){
 DataType=Type;
 switch(Type){
  case BOByte:	ElementSize=sizeof(SADBu8);break;
  case BOInteger:	ElementSize=sizeof(SADBu32);break;
  case BOFloat:	ElementSize=sizeof(SADBfl);break;
  case BODouble:	ElementSize=sizeof(SADBdfl);break;
  default:	ElementSize=1;//Default to single byte
 }
}

/// The the size in the nth dimension
/// @param D
///  Index of the desired dimension (starts at 0)
/// @return The size, or 0 if dimension doesn't exist
SADBu32 BlackboardObject::GetDimension(SADBu32 D){
 if(D>=Dimensionality)return 0;
 else return Dimensions[D];
}

/// Sets the size of the object by defining its dimensionality and its dimensions
/// @param D
///   Number of Dimensions
/// @param ...
///   Dimensions (Height, Width, etc)
/// @return 0 on success
int BlackboardObject::SetDimensionsf(SADBu32 D,...){
 Dimensionality=D;
 Dimensions=(SADBu32 *)realloc(Dimensions,D*sizeof(SADBu32));//Allocate array of dimensions
 if((Dimensions==NULL)&&(D!=0))return 1;
 TotalNumberOfElements=0;
 if(D>0){
  va_list DimensionsList;va_start(DimensionsList,D);//Setup variadic argument list
  TotalNumberOfElements=1;
  for(SADBu32 i=0;i<D;i++){//Set Dimensions
   Dimensions[i]=(SADBu32)va_arg(DimensionsList,int);
   TotalNumberOfElements*=Dimensions[i];
  }
  va_end(DimensionsList);
 }
 Data=realloc(Data,TotalNumberOfElements*ElementSize);//Allocate Memory Space to hold the object data
 if((Data==NULL)&&(TotalNumberOfElements!=0))return 2;
 return 0;
}

/// Sets the size of the object by defining its dimensionality and its dimensions, without allocating memory
/// @param D
///   Number of Dimensions
/// @param ...
///   Dimensions (Height, Width, etc)
/// @return 0 on success
int BlackboardObject::OnlySetDimensionsf(SADBu32 D,...){
 Dimensionality=D;
 Dimensions=(SADBu32 *)realloc(Dimensions,D*sizeof(SADBu32));//Allocate array of dimensions
 if(Dimensions==NULL)return 1;
 va_list DimensionsList;va_start(DimensionsList,D);//Setup variadic argument list
 TotalNumberOfElements=1;
 for(SADBu32 i=0;i<D;i++){//Set Dimensions
  Dimensions[i]=(SADBu32)va_arg(DimensionsList,int);
  TotalNumberOfElements*=Dimensions[i];
 }
 va_end(DimensionsList);
 return 0;
}

/// Sets the size of the object by defining its dimensionality and its dimensions
/// @param D
///   Number of Dimensions
/// @param Ds
///   list of Dimensions (Height, Width, etc)
/// @return 0 on success
int BlackboardObject::SetDimensions(SADBu32 D,SADBu32*Ds){
 Dimensionality=D;
 Dimensions=(SADBu32 *)realloc(Dimensions,D*sizeof(SADBu32));//Allocate array of dimensions
 if(Dimensions==NULL)return 1;
 TotalNumberOfElements=1;
 for(SADBu32 i=0;i<D;i++){//Set Dimensions
  Dimensions[i]=Ds[i];
  TotalNumberOfElements*=Dimensions[i];
 }
 Data=realloc(Data,TotalNumberOfElements*ElementSize);//Allocate Memory Space to hold the object data
 if(Data==NULL)return 2;
 return 0;
}


/// Remove oldest value from the blackboard
void BlackboardObject::RemoveOldestValue(void){
 if(Board!=NULL)Board->RemoveOldestValue(ID);
}


/// Discard earliest values and keep only the n most recent
/// @param nValues
///   Number of Values to be kept
void BlackboardObject::KeepOnlyLatestValues(int nValues){
 if(Board!=NULL)Board->KeepOnlyLatestValues(ID,nValues);
}



/// Get an object from the blackboard or create it if it doesn't already exist 
/// @param name
///  Name of the object
void BlackboardObject::CreateIfNonExistant(const char * name){
 GetByName(name);//Get the ID of the object, if it exists
 if(ID==0){//Object does not exist yet
  AddToBlackboard();//Create object on the blackboard
  SetName(name);
 }
}


/// Set Human Readable name of the object
/// @param name
///   Name
/// @return 0 on success
int BlackboardObject::SetName(const char * name){
 int Length=strlen(name)+1;
 Name=(SADBu8*)realloc(Name,Length);
 if(Name==NULL)return 1;
 memcpy(Name,name,Length);

 if(Board!=NULL)Board->SetObjectName(ID,(const char *)Name);
 return 0;
}

/// Get Human Readable name of the object from the blackboard
void BlackboardObject::GetName(void){
 Board->GetObjectName(ID,&Name);
}

/// Set Human Readable description of the object
/// @param desc
///   Description
/// @return 0 on success
int BlackboardObject::SetDescription(const char * desc){
 int Length=strlen(desc)+1;
 Description=(SADBu8*)realloc(Description,Length);
 if(Description==NULL)return 1;
 memcpy(Description,desc,Length);
 if(Board!=NULL)Board->SetObjectDescription(ID,(const char *)Description);
 return 0;
}

/// Get Human Readable description of the object from the blackboard
void BlackboardObject::GetDescription(void){
 Board->GetObjectDescription(ID,&Description);
}


/// Set the timestamp
/// @param time
///  Desired timestamp
void BlackboardObject::SetTimestamp(Timestamp * time){
 Time.Second=time->Second;
 Time.Millisecond=time->Millisecond;
}

/// Set the timestamp to the current time
void BlackboardObject::SetTimestampNow(void){
 Time.Now();
}

/// Print Human Readable information about the object
void BlackboardObject::Information(void){
 printf("Name: %s\n",Name);
 printf("ID: %lld\n",ID);
 printf("Timestamp: %lf\n",Time.Get());
 printf("Description: %s\n",Description);
 printf("Contains %d dimensions\n",Dimensionality);
 printf("Dimensions: ");
 for(SADBu32 i=0;i<Dimensionality;i++)printf("%d ",Dimensions[i]);
 printf("\n");

 switch(DataType){//Convert data at given memory location to proper type
  case BOByte:	printf("Data Type: Byte\n");break;
  case BOInteger:	printf("Data Type: Integer\n");break;
  case BOFloat:	printf("Data Type: Float\n");break;
  case BODouble:	printf("Data Type: Double\n");break;
  default:	printf("Data Type: Defaulted to Byte\n");
 }

 if(TotalNumberOfElements<100){
  printf("Raw Data: \n");
  Pointer p=(Pointer)Data;
  for(int i=0;i<TotalNumberOfElements;i++){
   switch(DataType){//Convert data at given memory location to proper type
    case BOByte:	printf("%03d\t",*((SADBu8*)p));	p+=1;break;
    case BOInteger:	printf("%d\t",	*((SADBu32*)p));p+=4;break;
    case BOFloat:	printf("%f\t",	*((SADBfl*)p));	p+=4;break;
    case BODouble:	printf("%lf\t",	*((SADBdfl*)p));p+=8;break;
    default:		printf("%03d\t",*((SADBu8*)p));	p+=1;
   }
  }  
  printf("\n");
 }

}




/// Get the content of the object as a string
/// @note this function assumes that the object contains a single string, for string tables, look a the SADBStringTable class
/// @return Pointer to the start of the string
const char * BlackboardObject::GetString(void){
 return (char*)Data;
}

/// Get the content of the object as a string
/// @note this function assumes that the object contains a single string, for string tables, look a the SADBStringTable class
/// @note This functions allocates memory, you mustt free the returned pointer yourself
/// @return Pointer to the start of the string
char * BlackboardObject::GetNewString(void){
 int l=strlen((const char*)Data);
 char * New=(char*)malloc((l+1)*sizeof(char));
 memcpy(New,Data,l);
 New[l]='\0';
 return New;
}




/// Get the value of a byte in the object at a given location
/// @param X Position in the first dimension (required as a first argument to va_start)
/// @param ... Position of the element in the other dimensions
SADBu8 BlackboardObject::GetByte(int X,...){
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,X);Pos[0]=X;
 for(int i=1;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 Pointer P=GetElementLocation(Pos);
 free(Pos);
 return *((SADBu8*)P);
}

/// Get the value of a boolean in the object at a given location
/// @note Booleans are stored as bytes (false is stored as 0, true as 1)
/// @param X Position in the first dimension (required as a first argument to va_start)
/// @param ... Position of the element in the other dimensions
bool BlackboardObject::GetBoolean(int X,...){
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,X);Pos[0]=X;
 for(int i=1;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 Pointer P=GetElementLocation(Pos);
 free(Pos);
 if(*((SADBu8*)P)==1)return true;
 else return false;
}

/// Get the value of an integer in the object at a given location
/// @param X Position in the first dimension (required as a first argument to va_start)
/// @param ...  Position of the element in the other dimensions
SADBu32 BlackboardObject::GetInteger(int X,...){
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,X);Pos[0]=X;
 for(int i=1;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 Pointer P=GetElementLocation(Pos);
 free(Pos);
 return *((SADBu32*)P);
}


/// Get the value of a float in the object at a given location
/// @param X Position in the first dimension (required as a first argument to va_start)
/// @param ... Position of the element in the other dimensions
SADBfl BlackboardObject::GetFloat(int X,...){
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,X);Pos[0]=X;
 for(int i=1;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 Pointer P=GetElementLocation(Pos);
 free(Pos);
 return *((SADBfl*)P);
}


/// Get the value of a double precision float in the object at a given location
/// @param X Position in the first dimension (required as a first argument to va_start)
/// @param ... Position of the element in the other dimensions
SADBdfl BlackboardObject::GetDouble(int X,...){
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,X);Pos[0]=X;
 for(int i=1;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 Pointer P=GetElementLocation(Pos);
 free(Pos);
 return *((SADBdfl*)P);
}


/// Set a string as the content of the object.
/// @note this function assumes that the object contains a single string, for string tables, look a the SADBStringTable class
/// @note Allocates the required memory
/// @param str Desired string
void BlackboardObject::SetString(const char * str){
 int l=strlen(str)+1;//+1 for the '\0'
 int e=SetDimensionsf(1,l);//allocate memory
 if(Data==NULL){
  printf("Error %d \n",e);
  return;
 }
 memcpy(Data,str,l);
}


/// Get a pointer to the data
/// @param str Desired string
Pointer BlackboardObject::GetElementLocation(int*Pos){
 //Debug
 //printf("Getting Pointer to:");
 //for(int d=0;d<(int)Dimensionality-1;d++)printf("%d,",Pos[d]);
 //printf("\n"); 
 
 Pointer p=(Pointer)Data;
 for(int d=0;d<(int)Dimensionality-1;d++){
  int D=Dimensions[d+1];//product of all the other dimensions after it
  for(int i=d+2;i<(int)Dimensionality;i++)D*=Dimensions[i];
   p+=Pos[d]*D*ElementSize;
 }
 return p+Pos[Dimensionality-1]*ElementSize;
}


/// Sets a string in an object
/// @param str
///   String
/// @param ...
///   Position of the element to receive the value (X,Y,Z,...), the last dimension stores the string
void BlackboardObject::SetStringElement(const char * str,...){
 if(ElementSize!=1)return;//Not bytes
 int l=strlen(str)+1;//+1 for the '\0'
 //Assume last dimension is the string
 int ml=Dimensions[Dimensionality-1];
 if(l>ml)l=ml;//Truncate string if required
 
 //Build index array
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,str);
 for(int i=0;i<(int)Dimensionality-1;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);

 memcpy((SADBu8*)GetElementLocation(Pos),str,l);
 //Cheap way, left for posterity
 //for(int i=0;i<l;i++){Pos[Dimensionality-1]=i;*((SADBu8*)GetElementLocation(Pos))=str[i];}
 free(Pos);
}

/// Get the value of a float in the object at a given location
/// @param X
///   Position in the first dimension (required as a first argument to va_start)
/// @param ...
///   Position of the element in the other dimensions
char * BlackboardObject::GetStringElement(int X,...){
 if(ElementSize!=1)return NULL;//Not bytes
 char*S=(char*)malloc(Dimensions[Dimensionality-1]+1);//Storage for the output (incl. \0)
 S[Dimensions[Dimensionality-1]]='\0';
 
 //Build index array
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,X);Pos[0]=X;
 for(int i=1;i<(int)Dimensionality-1;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);

 memcpy(S,(SADBu8*)GetElementLocation(Pos),Dimensions[Dimensionality-1]);
 
 return S;
}





/// Sets the value of an element in the object
/// @param Value Value to set
/// @param ... Position of the element to receive the value (X,Y,Z,...)
void BlackboardObject::SetByte(SADBu8 Value,...){
 if(DataType!=BOByte)Error("Blackboard Object","SetByte called with wrong data type");
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,Value);
 for(int i=0;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 *((SADBu8*)GetElementLocation(Pos))=Value;//Set the value
 free(Pos);
}

/// Sets the value of an element in the object
/// @note Booleans are stored as bytes (false is stored as 0, true as 1)
/// @param Value Value to set
/// @param ... Position of the element to receive the value (X,Y,Z,...)
void BlackboardObject::SetBoolean(bool Value,...){
 if(DataType!=BOByte)Error("Blackboard Object","SetBoolean called with wrong data type");
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,Value);
 for(int i=0;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 if(Value==true)*((SADBu8*)GetElementLocation(Pos))=1;
 else *((SADBu8*)GetElementLocation(Pos))=0;
 free(Pos);
}

/// Sets the value of an element in the object
/// @param Value Value to set
/// @param ... Position of the element to receive the value (X,Y,Z,...)
void BlackboardObject::SetInteger(SADBu32 Value,...){
 if(DataType!=BOInteger)Error("Blackboard Object","SetInteger called with wrong data type");
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,Value);
 for(int i=0;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 *((SADBu32*)GetElementLocation(Pos))=Value;
 free(Pos);
}

/// Sets the value of an element in the object
/// @param Value Value to set
/// @param ... Position of the element to receive the value (X,Y,Z,...)
void BlackboardObject::SetFloat(SADBfl Value,...){
 if(DataType!=BOFloat)Error("Blackboard Object","SetFloat called with wrong data type");
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,Value);
 for(int i=0;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 *((SADBfl*)GetElementLocation(Pos))=Value;
 free(Pos);
}

/// Sets the value of an element in the object
/// @param Value
///   Value to set
/// @param ...
///   Position of the element to receive the value (X,Y,Z,...)
void BlackboardObject::SetDouble(SADBdfl Value,...){
 if(DataType!=BODouble)Error("Blackboard Object","SetDouble called with wrong data type");
 int*Pos=(int*)malloc(Dimensionality*sizeof(int));
 va_list Position;va_start(Position,Value);
 for(int i=0;i<(int)Dimensionality;i++)Pos[i]=va_arg(Position,int);
 va_end(Position);
 *((SADBdfl*)GetElementLocation(Pos))=Value;
 free(Pos);
}



/// Initialize the object with data from a comma separated values (CSV) file
/// @note The entire CSV file gets loaded as the value of the object for a given timestamp
/// @note Uses the current time as the timestamp
/// @param Filename
///  Path to the CSV file
/// @param DataType
///  Type of the data in the CSV File
void BlackboardObject::CreateWithInitialValueFromCSV(const char * Filename,enum BODataType DataType){
 Timestamp Time;
 Time.Now();
 CreateWithInitialValueFromCSV(&Time,Filename,DataType,',');
}

/// Initialize the object with data from a comma separated values (CSV) file
/// @note The entire CSV file gets loaded as the value of the object for a given timestamp
/// @note Uses the current time as the timestamp
/// @param Filename
///  Path to the CSV file
/// @param DataType
///  Type of the data in the CSV File
/// @param Delimiter
///  Delimiter (normally ',')
void BlackboardObject::CreateWithInitialValueFromCSV(const char * Filename,enum BODataType DataType,char Delimiter){
 Timestamp Time;
 Time.Now();
 CreateWithInitialValueFromCSV(&Time,Filename,DataType,Delimiter);
}

/// Initialize the object with data from a comma separated values (CSV) file
/// @note The entire CSV file gets loaded as the value of the object for a given timestamp
/// @param Time
///  Pointer to a timestamp object
/// @param Filename
///  Path to the CSV file
/// @param DataType
///  Type of the data in the CSV File
/// @param Delimiter
///  Delimiter (normally ',')
void BlackboardObject::CreateWithInitialValueFromCSV(Timestamp * time,const char * Filename,enum BODataType DataType,char Delimiter){

 //Extract real file name (remove path and .csv)
 int fnl=strlen(Filename);
 int ls=0;
 for(int i=0;i<fnl;i++)if(Filename[i]=='/')ls=i+1;
 int onl=fnl-ls-4+1;
 char*on=(char*)malloc(onl+1);
 memcpy(on,Filename+ls,onl);on[onl-1]='\0';

 if(Board!=NULL)GetByName(on);
 if(ID==0){//not already on blackboard
  SetName(on);//Set the object name
  if(Board!=NULL)AddToBlackboard();
 }
 SetDataType(DataType);
 
 Time.Set(time->Second,time->Millisecond);

 free(on);//file is open discard filename


 FILE *f=fopen(Filename,"r");
 int c=0;//store the current character

 //Find number of rows and columns
 int nC=0;
 while((c!='\n')&&(c!=EOF)){
  c=fgetc(f);
  if(c==Delimiter)nC++;
 }
 if(nC>0)nC++;
 rewind(f);
 int nR=0;
 while(c!=EOF){
  c=fgetc(f);
  if((c=='\n'))nR++;
 }
 rewind(f);

 printf(" CSV File has %d columns and %d rows\n",nC,nR);

 if((nC>0)&&(nR>0))SetDimensionsf(2,nR,nC);//2D table
 else SetDimensionsf(0);//Vector


 char * Tmp=(char*)malloc(1000);//Max string length

 //Now start extracting data
 for(int y=0;y<nR;y++){

  for(int x=0;x<nC;x++){
   //find next ',' or '\n'
   c=0;
   int nd=0;
   while((c!=EOF)&&(c!='\n')&&(c!=Delimiter)){
    c=fgetc(f);
    nd++;
   }

   fseek(f,-nd,SEEK_CUR);
   fgets(Tmp,nd,f);
   fseek(f,1,SEEK_CUR);
   //printf("%d,%d => %s\n",x,y,Tmp);

   if(DataType==BOByte){
    SetByte(atoi(Tmp),y,x);
   }else if(DataType==BOInteger){
    SetInteger(atoi(Tmp),y,x);
   }else if(DataType==BOFloat){
    SetFloat((SADBfl)atof(Tmp),y,x);
   }else if(DataType==BODouble){
    SetDouble((SADBdfl)atof(Tmp),y,x);
   }



  }

 }

 if(Board!=NULL)Send();
 fclose(f);
}


/// Write the values of the object to a CSV file.  The first element of each line is the timestamp and the remaining cells of the line contain the value.
/// @note For the time being, each value is vectorized to fit on a line, dimensions information is discarded.  This may cause problems with objects of higher dimensions.
/// @note If the dimensions of the object changes over time, the lines of the CSV file will have different lengths in order to export the entire data.
/// @param Filename
///  Path to the CSV file
void BlackboardObject::ExportAllValuesToCSV(const char * Filename){
 Timestamp T;
 FILE *File=fopen(Filename,"w");
 while(1){
  GetValueAtNextTimestamp(&T);
  if((T.Second==Time.Second)&&(T.Millisecond==Time.Millisecond))break;
  T.Second=Time.Second;
  T.Millisecond=Time.Millisecond;
  //Output to file
  fprintf(File,"%lf",Time.Get());
  for(int i=0;i<TotalNumberOfElements;i++){
   if(DataType==BOByte)fprintf(File,",%d",GetByte(i));//writing it as an int
   if(DataType==BOInteger)fprintf(File,",%d",GetInteger(i));
   if(DataType==BOFloat)fprintf(File,",%f",GetFloat(i));
   if(DataType==BODouble)fprintf(File,",%lf",GetDouble(i));
  }
  fprintf(File,"\n");
 }
}




/// Initialize the object with data from a comma separated values (CSV) file
/// @note the first column of each row should contain a timestamp encoded as a double, each row of data will be loaded on the blackboard as a different value at a distinct timestamp
/// @param Filename
///  Path to the CSV file
/// @param DataType
///  Type of the data in the CSV File (other than the first column, which MUST be a double/float) 
void BlackboardObject::CreateWithValuesFromCSV(const char * Filename,enum BODataType DataType){
 CreateWithValuesFromCSV(Filename,DataType,',');
}



/// Initialize the object with data from a comma separated values (CSV) file
/// @note the first column of each row should contain a timestamp encoded as a double, each row of data will be loaded on the blackboard as a different value at a distinct timestamp
/// @param Filename
///  Path to the CSV file
/// @param DataType
///  Type of the data in the CSV File (other than the first column, which MUST be a double/float) 
/// @param Delimiter
///  Delimiter (normally ',')
void BlackboardObject::CreateWithValuesFromCSV(const char * Filename,enum BODataType DataType,char Delimiter){

 //Extract real file name (remove path and .csv)
 int fnl=strlen(Filename);
 int ls=0;
 for(int i=0;i<fnl;i++)if(Filename[i]=='/')ls=i+1;
 int onl=fnl-ls-4+1;
 char*on=(char*)malloc(onl+1);
 memcpy(on,Filename+ls,onl);on[onl-1]='\0';

 if(Board!=NULL)GetByName(on);
 if(ID==0){//not already on blackboard
  SetName(on);//Set the object name
  if(Board!=NULL)AddToBlackboard();
 }
 SetDataType(DataType);

 free(on);//file is open discard filename


 FILE *f=fopen(Filename,"r");
 int c=0;//store the current character

 //Find number of rows and columns
 int nC=0;
 while(c!='\n'){
  c=fgetc(f);
  if(c==',')nC++;
 }
 if(nC>0)nC++;
 rewind(f);
 int nR=0;
 while(c!=EOF){
  c=fgetc(f);
  if(c=='\n')nR++; 
 }
 rewind(f);

 //The first column is the timestamp, so use nC-1

 if(nC>1)SetDimensionsf(1,nC-1);
 else SetDimensionsf(0);


 char * Tmp=(char*)malloc(1000);//Max string length

 //Now start extracting data
 for(int y=0;y<nR;y++){//foreach line

  //extract up to the first ',', that will be the timestamp
  c=0;
  int nd=0;
  while(c!=EOF){
   c=fgetc(f);
   nd++;
   if((c==Delimiter)||(c=='\n'))break;
  }
  fseek(f,-nd,SEEK_CUR);
  fgets(Tmp,nd,f);
  fseek(f,1,SEEK_CUR);

   //Need to skip the \r or \n
   c=fgetc(f);
   if((c!='\n')&&(c!='\r'))fseek(f,-1,SEEK_CUR);


  double TS=atof(Tmp);
  Time.Set(TS);


  for(int x=0;x<nC-1;x++){
   //find next ',' or '\n'
   c=0;
   int nd=0;
   while(c!=EOF){
    c=fgetc(f);
    nd++;
    if((c==Delimiter)||(c=='\n')||(c=='\r'))break;
   }

   fseek(f,-nd,SEEK_CUR);
   fgets(Tmp,nd,f);
   fseek(f,1,SEEK_CUR);

   //Need to skip the \r or \n
   c=fgetc(f);
   if((c!='\n')&&(c!='\r'))fseek(f,-1,SEEK_CUR);

   if(DataType==BOByte){
    SetByte(atoi(Tmp),x);
   }else if(DataType==BOInteger){
    SetInteger(atoi(Tmp),x);
   }else if(DataType==BOFloat){
    SetFloat((SADBfl)atof(Tmp),x);
   }else if(DataType==BODouble){
    SetDouble((SADBdfl)atof(Tmp),x);
   }

  }

  if(Board!=NULL)Send();//Send this value

 }


 fclose(f);
}



/// Write the values of the object to a raw file.  Similar to the exportCSV function, but this one also stores information about the dimensions of objects.  The first element of each line is the timestamp, the second element is the data type, the third column is the number of dimensions, the next few columns are the dimensions, and the remaining cells contain the value.
/// @param Filename
///  Path to the output file
void BlackboardObject::ExportRaw(const char * Filename){
 Timestamp End;
 FILE *File=fopen(Filename,"w");
 
 //Get the latest timestamp
 if(GetLatestValue()!=0){//No values
  fclose(File);
  return;
 }
 End.Second=Time.Second;
 End.Millisecond=Time.Millisecond;



 Time.Second=0;
 Time.Millisecond=0;

 int TS=0;
 while(1){
  GetValueAtNextTimestamp(&Time);
  //Output to file
  fprintf(File,"%lf",Time.Get());
  fprintf(File,",%d",DataTypeEnumToInt(DataType));
  fprintf(File,",%d,",Dimensionality);
  for(SADBu32 i=0;i<Dimensionality;i++)fprintf(File,"%d,",Dimensions[i]);

  int Size=sizeof(SADBu8);
  if(DataType==BOInteger)Size=sizeof(SADBu32);
  if(DataType==BOFloat)Size=sizeof(SADBfl);
  if(DataType==BODouble)Size=sizeof(SADBdfl);
  fwrite(Data,Size,TotalNumberOfElements,File);
  
  fprintf(File,"\n");
  TS++;
  if((Time.Second==End.Second)&&(Time.Millisecond==End.Millisecond))break;//that was the last value to output
 }
 fclose(File);
 //printf("Dumped %d Values\n",TS);
}




/// Initialize the object with data from a raw data file (exported with exportraw)
/// @param Filename Path to the raw file
void BlackboardObject::ImportRaw(const char * Filename){
 ImportRaw(Filename,0,100);
}


/// Initialize the object with data from a raw data file (exported with exportraw)
/// @param Filename Path to the raw file
/// @param A Beginning Percentage
/// @param A End Percentage
void BlackboardObject::ImportRaw(const char * Filename,float A,float B){
 FILE *F=fopen(Filename,"r");
 //Check for errors
 if(F==NULL){printf("File %s does not exist\n",Filename);fclose(F);return; }
 fseek(F,0,SEEK_END);
 if(ftell(F)==0){printf("Empty File\n");fclose(F);return;  }
 rewind(F);

 //Temporary parameter storage
 double time;
 int type;
 SADBu32 nD;
 
 //Cont number of timestamps
 int nTS=0;
 while(++nTS){
  fscanf(F,"%lf,%d,%d,",&time,&type,&nD);
  enum BODataType datatype=DataTypeIntToEnum(type);
  int nValues=1;for(SADBu32 i=0;i<nD;i++){int t;fscanf(F,"%d,",&t);nValues*=t;}//Count total number of values
  //Determine Size of each element
  int size=sizeof(SADBu8);
  if(datatype==BOInteger)size=sizeof(SADBu32);
  if(datatype==BOFloat)size=sizeof(SADBfl);
  if(datatype==BODouble)size=sizeof(SADBdfl);
  fseek(F,nValues*size,SEEK_CUR);//Skip the actual value
  fseek(F,1,SEEK_CUR);//Skip newline at end of value
  if(fgetc(F)==EOF)break;//Done if we reached EOF
  else fseek(F,-1,SEEK_CUR);
 }
 rewind(F);
 printf("Object Contains %d timestamps\n",nTS);
 
 int Start=floor((A*nTS)/100);
 int Stop=ceil((B*nTS)/100);
 
 
 int TS=0;
 while(++TS){
  if(TS>Stop)break;//reached the end
  fscanf(F,"%lf,%d,%d,",&time,&type,&nD);
  Time.Set(time);
  SetDataType(DataTypeIntToEnum(type));
  SADBu32 *D=(SADBu32*)malloc(nD*sizeof(SADBu32));//Allocate memory to store dimensions
  for(SADBu32 i=0;i<nD;i++)fscanf(F,"%d,",&D[i]);//Get dimensions
  SetDimensions(nD,D);//Set dimensions
  free(D);//Free dimensions memory
  //Set the data size per element
  int Size=sizeof(SADBu8);
  if(DataType==BOInteger)Size=sizeof(SADBu32);
  if(DataType==BOFloat)Size=sizeof(SADBfl);
  if(DataType==BODouble)Size=sizeof(SADBdfl);
  fread(Data,Size,TotalNumberOfElements,F);//Read the content of the value
  if(TS>=Start)Send();//Send the object to the blackboard
  fseek(F,1,SEEK_CUR);//Skip newline at end of value
  if(fgetc(F)==EOF)break;//Done if we reached EOF
  else fseek(F,-1,SEEK_CUR);
 }
 fclose(F);
 //printf("Loaded %d Values\n",TS);
}



/// Write the values of the object to an ascii file.  Similar to the exportCSV function, but this one also stores information about the dimensions of objects.  The first element of each line is the timestamp, the second element is the data type, the third column is the number of dimensions, the next few columns are the dimensions, and the remaining cells contain the value.
/// @note Used to dump in the old format
/// @param Filename Path to the output file
void BlackboardObject::ExportAscii(const char * Filename){
 Timestamp End;
 FILE *File=fopen(Filename,"w");
 
 //Get the latest timestamp
 if(GetLatestValue()!=0){//No values
  fclose(File);
  return;
 }
 End.Second=Time.Second;
 End.Millisecond=Time.Millisecond;



 Time.Second=0;
 Time.Millisecond=0;

 int TS=0;
 while(1){
  GetValueAtNextTimestamp(&Time);
  //Output to file
  fprintf(File,"%lf",Time.Get());
  fprintf(File,",%d",DataTypeEnumToInt(DataType));
  fprintf(File,",%d",Dimensionality);
  for(SADBu32 i=0;i<Dimensionality;i++)fprintf(File,",%d",Dimensions[i]);

  Pointer p=(Pointer)Data;

  for(int i=0;i<TotalNumberOfElements;i++){
        if(DataType==BOByte)	fprintf(File,",%d",*((SADBu8*)p));
   else if(DataType==BOInteger)	fprintf(File,",%d",*((SADBu32*)p));
   else if(DataType==BOFloat)	fprintf(File,",%f",*((SADBfl*)p));
   else if(DataType==BODouble)	fprintf(File,",%lf",*((SADBdfl*)p));
   p+=ElementSize;//position of next element in memory
  }
  fprintf(File,"\n");
  TS++;
  if((Time.Second==End.Second)&&(Time.Millisecond==End.Millisecond))break;//that was the last value to output
 }
 fclose(File);
 //printf("Dumped %d Values\n",TS);
}




/// Initialize the object with data from an ascii file
/// @note This is to import legacy data
/// @param Filename
///  Path to the ascii file
void BlackboardObject::ImportAscii(const char * Filename){
 FILE *f=fopen(Filename,"r");
 printf("Importing %s\n",Filename);
 //Check to see if there's any content
  fseek(f,0,SEEK_END);
  printf("Importing %ld\n",ftell(f));
  if(ftell(f)==0){
   printf("Empty File\n");
   fclose(f);
   return;
  }
 rewind(f);
 
 int c=0;//store the current character
 int l;

 //Extract a line form the file
 char Temp[500];//Temporary Storage

 int TS=0;
 while(1){
  TS++;
  //Get Timestamp
  l=0;c=0;
  while(c!=','){
   c=fgetc(f);
   Temp[l]=(char)c;
   l++;
  }Temp[l-1]='\0';//The last char will be the ',', so replace by \0
  Time.Set(atof(Temp));//Set Timestamp

  //Get DataType
  l=0;c=0;
  while(c!=','){
   c=fgetc(f);
   Temp[l]=(char)c;
   l++;
  }Temp[l-1]='\0';//The last char will be the ',', so replace by \0
  SetDataType(DataTypeIntToEnum(atoi(Temp)));//Set DataType

  //Get number of dimensions
  l=0;c='a';
  while((c!=',')&&(c!='\n')&&(c!='\0')&&(c!=EOF)){
   c=fgetc(f);
   Temp[l]=(char)c;
   l++;
  }Temp[l-1]='\0';//The last char will be the ',', so replace by \0

  SADBu32 nD=atoi(Temp);
  SADBu32 *D=(SADBu32*)malloc(nD*sizeof(SADBu32));

  //Get dimensions
  for(SADBu32 i=0;i<nD;i++){

   l=0;c='a';
   while((c!=',')&&(c!='\n')&&(c!='\0')&&(c!=EOF)){
    c=fgetc(f);
    Temp[l]=(char)c;
    l++;
   }Temp[l-1]='\0';//The last char will be the ',', so replace by \0

   D[i]=(SADBu32)atoi(Temp);
  }
  SetDimensions(nD,D);//Set dimensions
  free(D);

  //Set the data
  Pointer p=(Pointer)Data;//Pointer to the start of the data

  for(int i=0;i<TotalNumberOfElements;i++){
   l=0;c='a';
   while((c!=',')&&(c!='\n')&&(c!='\0')&&(c!=EOF)){
    c=fgetc(f);
    //if(c==EOF)printf("EOF\n");
    Temp[l]=(char)c;
    l++;
   }Temp[l-1]='\0';//The last char will be the ',', so replace by \0

   //Actually set the element
        if(DataType==BOByte)*((SADBu8*)p)=(SADBu8)atoi(Temp);
   else if(DataType==BOInteger)*((SADBu32*)p)=(SADBu32)atoi(Temp);
   else if(DataType==BOFloat)*((SADBfl*)p)=(SADBfl)atof(Temp);
   else if(DataType==BODouble)*((SADBdfl*)p)=(SADBdfl)atof(Temp);
   p+=ElementSize;//position of next element in memory
  }
  Send();

  c=fgetc(f);
  if(c==EOF)break;
  else fseek(f,-1,SEEK_CUR);

 }
 fclose(f);
 //printf("Loaded %d Values\n",TS);
}


