#include "SADB.h"

/// Receive data from the socket
/// @param[in] Socket Socket to use
/// @param[out] Buffer Pointer to memory location where data read is to be stored
/// @param[in] MaxLength Maximum data to be read
/// @return Number of bytes read
int SADB::SocketReceive(int Socket,unsigned char * Buffer,int MaxLength){
 int length=recv(Socket,(char*)Buffer,MaxLength,0);
 return length;
}

/// Receive a complete Packet from the socket
/// @param[in] Socket Socket to use
/// @param[out] Buffer Pointer to memory location where data read is to be stored
/// @param[in] Length Maximum data to be read
/// @return Number of bytes read
int SADB::SocketReceiveExactly(int Socket,unsigned char * Buffer,int Length){
 //Block until a complete packet has been received
 int Received=0;
 while(Received<Length){
  int l=recv(Socket,(char*)Buffer,Length-Received,0);
  Received+=l;
  Buffer+=l;
  if(l==0)return l;
 }
 return Received;
}

/// Send Data
/// @param[in] Socket Socket to use
/// @param[in] Buffer Pointer to memory location where data to be sent is stored
/// @param[in] Length Number of bytes to be sent
int SADB::SocketSend(int Socket,unsigned char * Buffer,size_t Length){
 size_t Sent=0;
 while(Sent<Length){
  int s=send(Socket,(const char *)Buffer,Length-Sent,0);
  if(s==SOCKET_ERROR)return 1;
  Sent+=s;
  Buffer+=s;
 }
 return 0;
}

/// Send a string
/// @param[in] Socket Socket to use
/// @param[out] Buffer Pointer to memory location where data to be sent is stored
int SADB::SocketSend(int Socket,unsigned char * Buffer){
 return SocketSend(Socket,Buffer,strlen((const char*)Buffer));
}

/// Send a Formatted String
/// @param[in] Socket Socket to use
/// @param[out] format Format string
/// @param[in] ... Additional variables
int SADB::SocketSendf(int Socket,unsigned char *format,...){
 va_list args;
 va_start(args,format);
 unsigned char TMPBuffer[MAXFORMATTEDLENGTH];
 vsprintf((char*)TMPBuffer,(const char*)format,args);
 return SocketSend(Socket,&TMPBuffer[0]);
}



SADB::Network::Network(void){
#ifdef WINDOWS
 WSADATA wsaData;
 if(WSAStartup(MAKEWORD(2,2),&wsaData)!=0)Error("Network","WSAStartup failed");
#endif
 pthread_mutex_init(&Mutex,NULL);
}
SADB::Network::~Network(void){Close();}

/// Close a Socket
void SADB::Network::Close(void){
#ifdef POSIX
 close(Socket);
#endif
#ifdef WINDOWS
 closesocket(Socket);
#endif
}

/// Open a server socket
/// @param[in] port Port to use
void SADB::Network::Server(int port,int MaxQueue){
 Port=port;
 struct addrinfo Requested;
 memset(&Requested,0x00,sizeof(Requested));
 Requested.ai_family=AF_UNSPEC;//IPv4 or IPv6
 Requested.ai_socktype=SOCK_STREAM;
 Requested.ai_flags=AI_PASSIVE;//Auto Determine IP
 char * PortStr=(char *)malloc(10);
 sprintf(PortStr,"%d",Port);
 if(getaddrinfo(NULL,PortStr,&Requested,&AddressInfo)!=0)Error("Network","Cannot Get Address Information");
 if((Socket=socket(AddressInfo->ai_family,AddressInfo->ai_socktype,AddressInfo->ai_protocol))==INVALID_SOCKET)Error("Network","Cannot Open Socket");
 if(bind(Socket,AddressInfo->ai_addr,AddressInfo->ai_addrlen)==-1)Error("Network","Cannot Bind To Socket");
 if(listen(Socket,MaxQueue)==-1)Error("Network","Cannot Listen on Socket");
 free(PortStr);
 const char flag=1;setsockopt(Socket,IPPROTO_TCP,TCP_NODELAY,&flag,sizeof(flag));
}


/// Wait for a connection
int SADB::Network::WaitForConnection(void){
 struct sockaddr_storage NewAddress;
 socklen_t AddressSize=sizeof(NewAddress);
 int NewSocketFD=accept(Socket,(struct sockaddr *)&NewAddress,&AddressSize);//Wait for connection
 if(NewSocketFD==-1)Error("Network","Socket Error");
 const char flag=1;setsockopt(Socket,IPPROTO_TCP,TCP_NODELAY,&flag,sizeof(flag));
 return NewSocketFD;
}

/// Connect to a server
/// @param[in] address Address to connect to
/// @param[in] port Port to use
int SADB::Network::ConnectTo(const char * address,int port){
 Port=port;
 struct addrinfo Requested;
 memset(&Requested,0x00,sizeof(Requested));
 Requested.ai_family=AF_UNSPEC;//IPv4 or IPv6
 Requested.ai_socktype=SOCK_STREAM;
 char * PortStr=(char *)malloc(10);
 sprintf(PortStr,"%d",Port);
 if(getaddrinfo(address,PortStr,&Requested,&AddressInfo)!=0)return 1;//Error("Network","Cannot Get Address Information");
 if((Socket=socket(AddressInfo->ai_family,AddressInfo->ai_socktype,AddressInfo->ai_protocol))==INVALID_SOCKET)return 3;//Error("Network","Cannot Open Socket");
 if(connect(Socket,AddressInfo->ai_addr,AddressInfo->ai_addrlen)==-1)return 3;//Error("Network","Cannot Connect to Socket");
 const char flag=1;setsockopt(Socket,IPPROTO_TCP,TCP_NODELAY,&flag,sizeof(flag));
 return 0;
}

/// Receive data from the socket
/// @param[in] Socket Socket to use
/// @param[out] Buffer Pointer to memory location where data read is to be stored
/// @param[in] MaxLength Maximum data to be read
/// @return Number of bytes read
int SADB::Network::Receive(unsigned char * Buffer,int MaxLength){
 pthread_mutex_lock(&Mutex);
 int i=SocketReceive(Socket,Buffer,MaxLength);
 pthread_mutex_unlock(&Mutex);
 return i;
}

/// Receive a complete Packet from the socket
/// @param[in] Socket Socket to use
/// @param[out] Buffer Pointer to memory location where data read is to be stored
/// @param[in] Length Maximum data to be read
/// @return Number of bytes read
int SADB::Network::ReceiveExactly(unsigned char * Buffer,int Length){
 pthread_mutex_lock(&Mutex);
 int i=SocketReceiveExactly(Socket,Buffer,Length);
 pthread_mutex_unlock(&Mutex);
 return i;
} 

/// Send a string
/// @param[in] Socket Socket to use
/// @param[out] Buffer Pointer to memory location where data to be sent is stored
int SADB::Network::Send(unsigned char * Buffer){
 pthread_mutex_lock(&Mutex);
 int r=SocketSend(Socket,Buffer);
 pthread_mutex_unlock(&Mutex);
 return r;
}

/// Send Data
/// @param[in] Socket Socket to use
/// @param[in] Buffer Pointer to memory location where data to be sent is stored
/// @param[in] Length Number of bytes to be sent
int SADB::Network::Send(unsigned char * Buffer,int Length){
 pthread_mutex_lock(&Mutex);
 int r=SocketSend(Socket,Buffer,Length);
 pthread_mutex_unlock(&Mutex);
 return r;
}

/// Send a Formatted String
/// @param[in] Socket Socket to use
/// @param[out] format Format string
/// @param[in] ... Additional variables
int SADB::Network::Sendf(unsigned char *format,...){
 va_list args;
 va_start(args,format);
 unsigned char TMPBuffer[MAXFORMATTEDLENGTH];
 vsprintf((char *)TMPBuffer,(const char*)format,args);
 pthread_mutex_lock(&Mutex);
 int r=SocketSend(Socket,TMPBuffer);
 pthread_mutex_unlock(&Mutex);
 return r;
}




















