#include "../SADB.h"
using namespace SADB;

/// Constructor, initializes object
/// @see SADBStringTable::Initialize
SADBStringTable::SADBStringTable(void){Initialize();}

/// Destructor, deinitializes object
/// @see SADBStringTable::Free
SADBStringTable::~SADBStringTable(void){Free();}

/// Initialize Model
void SADBStringTable::Initialize(void){
 Name=NULL;
}

/// Free all allocated memory
void SADBStringTable::Free(void){
 if(Name!=NULL)free(Name);
 for(int x=0;x<nCols;x++)for(int y=0;y<nRows;y++){
  int i=GetIndex(x,y);
  Objects[i].Free();
 }
 free(Objects);
}

//private function
int SADBStringTable::GetIndex(int X,int Y){
 return Y*nCols+X;
}

/// Get a stringtable object from the SADB server
/// @note Just a wrapper for SetName, to be consistent with the GetByName function of standard blackboard objects.
/// @param name
///  Desired name
void SADBStringTable::GetByName(const char * name){
 SetName(name);
}

/// Set the size fo the string table (always 2D)
/// @note Must set blackboard before allocating memory
/// @param X
///  Number of columns
/// @param Y
///  Number of Rows
void SADBStringTable::SetDimensions(int X, int Y){
 nCols=X;
 nRows=Y;
 Object.SetDataType(BOInteger);
 Object.SetDimensionsf(3,X,Y,2);

 //Allocate objects and initialize them
 Objects=(BlackboardObject*)malloc(X*Y*sizeof(BlackboardObject));
 for(int i=0;i<X*Y;i++)Objects[i].Initialize();

 //Set the names
 char Tmp[1000];
 for(int x=0;x<X;x++)for(int y=0;y<Y;y++){
  int i=GetIndex(x,y);
  sprintf(Tmp,"%s-%dx%d",Name,x,y);
  Objects[i].SetBlackboard(Board);
  Objects[i].GetByName(Tmp);
  if(Objects[i].ID==0){
   Objects[i].SetName(Tmp);
   Objects[i].AddToBlackboard();
  }
  ((SADBu64*)Object.Data)[i]=Objects[i].ID;
 }
 Object.Send();
}

/// Get the latest version of the table from the board
void SADBStringTable::GetLatestValues(void){
 Object.GetLatestValue();

 int X=Object.Dimensions[0];
 int Y=Object.Dimensions[1];
 if((X!=nCols)||(Y!=nRows)){//The table changed Shape, reload entirely, not optimal, but safe
  //Deallocate Memory
  for(int x=0;x<nCols;x++)for(int y=0;y<nRows;y++){
   int i=GetIndex(x,y);
   Objects[i].Free();
  }
  free(Objects);
  //Reallocate memory
  Objects=(BlackboardObject*)malloc(X*Y*sizeof(BlackboardObject));
  for(int i=0;i<X*Y;i++)Objects[i].Initialize();
  nCols=X;
  nRows=Y;
  //Set IDs
  for(int x=0;x<nCols;x++)for(int y=0;y<nRows;y++){
   int i=GetIndex(x,y);
   Objects[i].ID=((SADBu64*)Object.Data)[i];
  }
 }

 //Load latest values
 for(int x=0;x<X;x++)for(int y=0;y<nRows;y++)
  GetLatestValue(x,y);//update string at position x,y

}


/// Get the latest value of a cell
/// @param X
///  Column index
/// @param Y
///  Row index
void SADBStringTable::GetLatestValue(int X,int Y){
 int i=GetIndex(X,Y);
 Objects[i].GetLatestValue();
}

/// Set a string in the table
/// @param X
///  Column position
/// @param Y
///  Row Position
/// @param Str
///  String to be set
void SADBStringTable::SetString(int X,int Y,const char * Str){
 int i=GetIndex(X,Y);
 Objects[i].SetString(Str);
 Objects[i].Send();
}

/// Get a string in the table
/// @param X
///  Column position
/// @param Y
///  Row Position
/// @return Pointer to the string (don't forget to free it) 
///  String to be set
const char * SADBStringTable::GetString(int X,int Y){
 int i=GetIndex(X,Y);
 return Objects[i].GetString();
}


/// Get a string in the table
/// @param X
///  Column position
/// @param Y
///  Row Position
/// @return Pointer to the string (don't forget to free it) 
///  String to be set
char * SADBStringTable::GetNewString(int X,int Y){
 int i=GetIndex(X,Y);
 return Objects[i].GetNewString();
}

/// Passtrough function that sets the name of the underlying object
/// @param name
///  Name of the object
void SADBStringTable::SetName(const char * name){
 int Length=strlen(name)+1;
 Name=(SADBu8*)realloc(Name,Length);
 memcpy(Name,name,Length);
 Object.GetByName(name);
 if(Object.ID==0){
  Object.SetName(name);
  Object.AddToBlackboard();
 }
}


/// Attach the string table to a blackboard
/// @param board
///  Blackboard to be used
void SADBStringTable::SetBlackboard(Blackboard * board){
 Board=board;
 Object.SetBlackboard(board);
}




/// Passtrough function that sets the description of the underlying object
/// @param desc
///  Description
void SADBStringTable::SetDescription(const char * desc){
 Object.SetDescription(desc);
}

/// Load a csv file containing strings
void SADBStringTable::LoadCSV(const char * Filename){
  //Extract real file name (remove path and .csv)
 int fnl=strlen(Filename);
 int ls=0;
 for(int i=0;i<fnl;i++)if(Filename[i]=='/')ls=i+1;
 int onl=fnl-ls-4+1;
 char*on=(char*)malloc(onl+1);
 memcpy(on,Filename+ls,onl);on[onl-1]='\0';

 SetName(on);
 free(on);

 
 FILE *f=fopen(Filename,"r");
 int c=0;//store the current character

 //Find number of rows and columns
 int nC=0;
 while(c!='\n'){
  c=fgetc(f);
  if(c==','){
   nC++;
  }
 }
 if(nC>0)nC++;
 rewind(f);
 int nR=0;
 while(c!=EOF){
  c=fgetc(f);
  if((c=='\n')||(c=='\r'))nR++; 
 }
 rewind(f);

 SetDimensions(nC,nR);

  char * Tmp=(char*)malloc(1000);//Max string length

 //Now start extracting data
 for(int y=0;y<nR;y++){
  for(int x=0;x<nC;x++){
   //find next ',' or '\n'
   c=0;
   int nd=0;
   while(c!=EOF){
    c=fgetc(f);
    nd++;
    if((c==',')||(c=='\n'))break;
   }

   fseek(f,-nd,SEEK_CUR);
   fgets(Tmp,nd,f);
   fseek(f,1,SEEK_CUR);

   SetString(x,y,Tmp);
  }
 }



}
