#include "SADB.hpp"

///Wait for a response from the server, don't care about the contents
/// @return 0 on success, 1 if client disconnected
int SADBWaitForResponse(int Socket){
 u8 CMD;u64 ID,Argument;
 u32 DataLength;u8 *Data=NULL;
 u64 Second;u16 Millisecond;
 int r=SADBReceive(Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,&Data);
 if(Data!=NULL)free(Data);//Just in case
 return r;
}


///Receive a packet and keep the string
/// @param String Pointer to a string to store result.  The memory will be allocated as needed.
/// @return 0 on success, 1 if client disconnected
int SADBReceiveString(int Socket,u8**String){
 u8 CMD;u64 ID,Argument;
 u32 DataLength;
 u64 Second;u16 Millisecond;
 int r=SADBReceive(Socket,&CMD,&ID,&Argument,&Second,&Millisecond,&DataLength,String);
 return r;
}

///Receive a packet and keep the ID
/// @param String Pointer to a string to store result.  The memory will be allocated as needed.
/// @return 0 on success, 1 if client disconnected
int SADBReceiveID(int Socket,u64*ID){
 u8 CMD;u64 Argument;
 u32 DataLength;u8 *Data=NULL;
 u64 Second;u16 Millisecond;
 int r=SADBReceive(Socket,&CMD,ID,&Argument,&Second,&Millisecond,&DataLength,&Data);
 if(Data!=NULL)free(Data);//Just in case
 return r;
}




///Receive an SADB Packet
/// @param Header Pointer to the memory location of the header
/// @param CMD Command to send
/// @param OUID Object Unique ID
/// @param Argument Optional argument
/// @param Second Timestamp of the object
/// @param Millisecond Timestamp of the object
/// @param DataLength Length of the data segment
/// @param Data Pointer to some memory to store the data, will be allocated
/// @return 0 on success, 1 if client disconnected
int SADBReceive(int Socket,u8 *CMD,u64 *OUID,u64 *Argument,u64 *Second,u16 *Millisecond,u32 *DataLength,u8 ** Data){
 u8 Header[SADBHeaderLength];
 if(SocketReceiveExactly(Socket,Header,SADBHeaderLength)==0){return 1;}
 if((Header[0x00]!=0x5A)||(Header[0x01]!=0xDB)){return 2;}//Sanity check
 //Extract info
 *CMD=Header[0x02];
 *OUID=*((u64*)(Header+0x03));
 *Argument=*((u64*)(Header+0x0B));
 *Second=*((u64*)(Header+0x13));
 *Millisecond=*((u16*)(Header+0x1B));
 *DataLength=*((u32*)(Header+0x1D));
 if(*DataLength>0){
  *Data=(u8 *)malloc(*DataLength);
  if(SocketReceiveExactly(Socket,*Data,*DataLength)==0)return 1;
 }
 return 0;
}


///Receive an SADB Packet
/// @param Header Pointer to the memory location of the header
/// @param CMD Command to send
/// @param OUID Object Unique ID
/// @param Argument Optional argument
/// @param Second Timestamp of the object
/// @param Millisecond Timestamp of the object
/// @param DataLength Length of the data segment
/// @param Data Pointer to some memory to store the data, will be reallocated
/// @return 0 on success, 1 if client disconnected
int SADBReceive2(int Socket,u8 *CMD,u64 *OUID,u64 *Argument,u64 *Second,u16 *Millisecond,u32 *DataLength,u8 ** Data){
 u8 Header[SADBHeaderLength];
 if(SocketReceiveExactly(Socket,Header,SADBHeaderLength)==0){return 1;}
 if((Header[0x00]!=0x5A)||(Header[0x01]!=0xDB)){return 2;}//Sanity check
 //Extract info
 *CMD=Header[0x02];
 *OUID=*((u64*)(Header+0x03));
 *Argument=*((u64*)(Header+0x0B));
 *Second=*((u64*)(Header+0x13));
 *Millisecond=*((u16*)(Header+0x1B));
 *DataLength=*((u32*)(Header+0x1D));
 if(*DataLength>0){
  *Data=(u8 *)realloc(*Data,*DataLength);
  if(SocketReceiveExactly(Socket,*Data,*DataLength)==0)return 1;
 }
 return 0;
}



/// Generate an SADB packet Header
/// @param Header Pointer to the memory location of the header
/// @param CMD Command to send
/// @param OUID Object Unique ID
/// @param Argument Optional argument
/// @param Second Timestamp of the object
/// @param Millisecond Timestamp of the object
/// @param DataLength Length of the data segment
void GenerateSADBHeader(u8* Header,u8 CMD,u64 OUID,u64 Argument,u64 Second,u16 Millisecond,u32 DataLength){
 Header[0x00]=0x5A;Header[0x01]=0xDB;//SADB Marker
 Header[0x02]=CMD;//Command
 *((u64*)(Header+0x03))=OUID;// Object ID 1
 *((u64*)(Header+0x0B))=Argument;// argument
 *((u64*)(Header+0x13))=Second;// Timestamp.Second
 *((u16*)(Header+0x1B))=Millisecond;// Timestamp.Millisecond
 *((u32*)(Header+0x1D))=DataLength;// Data Length
}



/// Send an SADB packet
/// @param Socket Target Socket
/// @param CMD Command to send
/// @param OUID Object Unique ID
/// @param Argument Optional argument
/// @param Second Timestamp of the object
/// @param Millisecond Timestamp of the object
/// @param DataLength Length of the data segment
/// @param Data Pointer to the data
void SendSADBPacket(int Socket,u8 CMD,u64 OUID,u64 Argument,u64 Second,u16 Millisecond,u32 DataLength,u8 * Data){
 //Cork the socket
 int flag=1;setsockopt(Socket,SOL_TCP,TCP_CORK,&flag,sizeof(flag));
 //Generate and Send the header
 u8 Header[SADBHeaderLength];
 GenerateSADBHeader(Header,CMD,OUID,Argument,Second,Millisecond,DataLength);
 SocketSend(Socket,Header,SADBHeaderLength);
 //Send the data
 if(DataLength!=0)SocketSend(Socket,Data,DataLength);
 //Uncork the socket
 flag=0;setsockopt(Socket,SOL_TCP,TCP_CORK,&flag,sizeof(flag));
}

/// Send an SADB packet with 2 data blocks
/// @param Socket Target Socket
/// @param CMD Command to send
/// @param OUID Object Unique ID
/// @param Argument Optional argument
/// @param Second Timestamp of the object
/// @param Millisecond Timestamp of the object
/// @param DataLengthA Length of the first data segment
/// @param DataA Pointer to the first data location
/// @param DataLengthB Length of the second data segment
/// @param DataB Pointer to the sencond data location
void SendSADBPacket(int Socket,u8 CMD,u64 OUID,u64 Argument,u64 Second,u16 Millisecond,u32 DataLengthA,u8 * DataA,u32 DataLengthB,u8 * DataB){
 //Cork the socket
 int flag=1;setsockopt(Socket,SOL_TCP,TCP_CORK,&flag,sizeof(flag));
 //Generate and Send the header
 u8 Header[SADBHeaderLength];
 GenerateSADBHeader(Header,CMD,OUID,Argument,Second,Millisecond,DataLengthA+DataLengthB);
 SocketSend(Socket,Header,SADBHeaderLength);
 if(DataLengthA!=0)SocketSend(Socket,DataA,DataLengthA);//First Block
 if(DataLengthB!=0)SocketSend(Socket,DataB,DataLengthB);//Second Block
 //Uncork the socket
 flag=0;setsockopt(Socket,SOL_TCP,TCP_CORK,&flag,sizeof(flag));
}
