#include "SADB.hpp"

/// Constructor, initializes object
/// @see Blackboard::Initialize
SADBCategory::SADBCategory(void){Initialize();}

/// Destructor, deinitializes object
/// @see Blackboard::Free
SADBCategory::~SADBCategory(void){Free();}

/// Initializes object
void SADBCategory::Initialize(void){
 Name=NULL;Description=NULL;
 ObjectIDs=NULL;nObjects=0;
 pthread_mutex_init(&InfoMutex,NULL);
 pthread_mutex_init(&DataAllocationMutex,NULL);
}

/// Frees allocated memory
void SADBCategory::Free(void){
 if(ObjectIDs!=NULL)free(ObjectIDs);
 if(Name!=NULL)free(Name);
 if(Description!=NULL)free(Description);
}


/// Set Human Readable name of the category
/// @param name
///   Name
/// @return 0 on success
int SADBCategory::SetName(const char * name){
 pthread_mutex_lock(&InfoMutex);
 int Length=strlen(name);
 if(Name!=NULL)free(Name);
 Name=(u8*)malloc(Length+1);//+1 to include '\0'
 if(Name==NULL)return 1;
 memcpy(Name,name,Length+1);
 pthread_mutex_unlock(&InfoMutex);
 return 0;
}

/// Set Human Readable description of the object
/// @param desc
///   Description
/// @return 0 on success
int SADBCategory::SetDescription(const char * desc){
 pthread_mutex_lock(&InfoMutex);
 int Length=strlen(desc);
 if(Description!=NULL)free(Description);
 Description=(u8*)malloc(Length+1);//+1 to include '\0'
 if(Description==NULL)return 1;
 memcpy(Description,desc,Length+1);
 pthread_mutex_unlock(&InfoMutex);
 return 0;
}







/// Add an object to this category
/// @param ID
///   ID of the object to add into category
/// @return 0 on success
int SADBCategory::AddObject(u64 ID){
 pthread_mutex_lock(&DataAllocationMutex);
  if(nObjects==0){//Blank Category
   nObjects++;
   ObjectIDs=(u64 *)realloc(ObjectIDs,nObjects*sizeof(u64));//Allocate more memory
   if(ObjectIDs==NULL)return 1;
   ObjectIDs[nObjects-1]=ID;
  }else{
   int AlreadyMember=0;
   for(int i=0;i<nObjects;i++)if(ObjectIDs[i]==ID)AlreadyMember=1;
   if(AlreadyMember==1){
    //printf("Already in category, not readding\n");   
   }else if(ID<ObjectIDs[nObjects-1]){//Need to reorder
    u64 *OldList=ObjectIDs;//Keep old list to copy from
    //Reallocate list
    ObjectIDs=NULL;
    nObjects++;
    ObjectIDs=(u64 *)malloc(nObjects*sizeof(u64));//Allocate more memory
    if(ObjectIDs==NULL)return 1;
    //Find New Position
    int Position;
    for(Position=nObjects-2;Position>0;Position--)
     if(ID>OldList[Position-1])break;
    ObjectIDs[Position]=ID;
    for(int i=0;i<Position;i++)ObjectIDs[i]=OldList[i];
    for(int i=Position+1;i<nObjects;i++)ObjectIDs[i]=OldList[i-1];
    free(OldList);
   }else{//Simply Append to the list
    nObjects++;
    ObjectIDs=(u64 *)realloc(ObjectIDs,nObjects*sizeof(u64));//Allocate more memory
    if(ObjectIDs==NULL)return 1;
    ObjectIDs[nObjects-1]=ID;
   }
  }
 pthread_mutex_unlock(&DataAllocationMutex);
 return 0;
}

/// Remove an object from this category
/// @param ID
///   ID of the object to be removed from category
/// @return 0 on success
int SADBCategory::RemoveObject(u64 ID){
 pthread_mutex_lock(&DataAllocationMutex);
  int Position=-1;
  for(int i=0;i<nObjects;i++)if(ObjectIDs[i]==ID){Position=i;break;}
  if(Position==-1)return 1;
  //Create a new list identical to ObjectIDs
  u64 * Copy=(u64 *)malloc((nObjects-1)*sizeof(u64));
  if(Copy==NULL)return 2;
  memcpy(Copy,ObjectIDs,nObjects);
  //Reallocate memory and copy IDs back into ObjectIDs
  ObjectIDs=(u64 *)realloc(ObjectIDs,(nObjects-1)*sizeof(u64));
  if(ObjectIDs==NULL)return 3;
  for(int j=0;j<Position;j++)ObjectIDs[j]=Copy[j];
  for(int j=Position+1;j<nObjects;j++)ObjectIDs[j-1]=Copy[j];
  free(Copy);
  nObjects--;
 pthread_mutex_unlock(&DataAllocationMutex);
 return 0;
}
