#include "SADB.hpp"

/// Constructor, initializes object
/// @see SADBObject::Initialize
SADBObject::SADBObject(void){Initialize();}

/// Destructor, deinitializes object
/// @see SADBObject::Free
SADBObject::~SADBObject(void){Free();}

/// Initializes object
void SADBObject::Initialize(void){
 Name=NULL;Description=NULL;
 Values=NULL;nValues=0;
 pthread_mutex_init(&InfoMutex,NULL);
 pthread_mutex_init(&DataAllocationMutex,NULL);
}

/// Frees allocated memory
void SADBObject::Free(void){
// printf("Removing %d %s\n",ID,Name);
 if(Name!=NULL)free(Name);Name=NULL;
 if(Description!=NULL)free(Description);Description=NULL;
 ID=0;
 if(Values!=NULL){
  for(int i=0;i<nValues;i++){Values[i]->Free();free(Values[i]);}
  free(Values);Values=NULL;
  nValues=0;
 }
 pthread_mutex_destroy(&InfoMutex);
 pthread_mutex_destroy(&DataAllocationMutex);
}

/// Add a new value
void SADBObject::NewValue(u32 Dimensionality,u32*Dimensions,u8 DataType,u8*Data,u64 Second, u16 Millisecond){
 //printf("Adding New Value to %s with time: %ld.%d\n",Name,Second,Millisecond);
 SADBObjectStorage * NewValue;
 NewValue=(SADBObjectStorage *)malloc(sizeof(SADBObjectStorage));
 NewValue->Initialize();
 NewValue->SetDataType(DataTypeIntToEnum(DataType));
 NewValue->SetDimensions(Dimensionality,Dimensions);
 NewValue->Time.Second=Second;
 NewValue->Time.Millisecond=Millisecond;
 memcpy(NewValue->Data,Data,NewValue->TotalNumberOfElements*NewValue->ElementSize);//copy data to new value

 pthread_mutex_lock(&DataAllocationMutex);
 AllocateValue();
 int pos;
 //Find proper position in the list
 if(nValues==1){//First value in the list
  pos=0;
 }else{
  pos=0;//default to 0
  for(int i=nValues-2;i>=0;i--){//Find where the new value goes
   //printf("%ld.%d\n",Values[i]->Time.Second,Values[i]->Time.Millisecond);
   if(TimestampDifference(Values[i]->Time.Second,Values[i]->Time.Millisecond,Second,Millisecond)>0){
    pos=i+1;
    break;
   }
  }
  if(pos!=nValues-1)//if not at the end of the list
   for(int i=nValues-1;i>pos;i--)Values[i]=Values[i-1];//Shift list to make space for the new value
 }
 Values[pos]=NewValue;//Insert the new value
 pthread_mutex_unlock(&DataAllocationMutex);
}

/// Allocate memory for a new value
/// @return 0 on success
int SADBObject::AllocateValue(void){
 nValues++;
 Values=(SADBObjectStorage **)realloc(Values,nValues*sizeof(SADBObjectStorage *));
 if(Values==NULL)return 1;
 return 0;
}

/// Remove a value
/// @param index
///  Index of the value to be removed
/// @return 0 on success
int SADBObject::RemoveValue(int index){
 pthread_mutex_lock(&DataAllocationMutex);
 SADBObjectStorage * RemovedValue=Values[index];
 SADBObjectStorage ** OldValues=Values;//In case something goes wrong
 nValues--;
 Values=NULL;
 Values=(SADBObjectStorage **)malloc(nValues*sizeof(SADBObjectStorage *));
 if(Values==NULL){
  Values=OldValues;
  nValues++;
  pthread_mutex_unlock(&DataAllocationMutex);
  return 1;
 }
 //memcpy(Values[0],OldValues[0],(index-1)*sizeof(SADBObjectStorage *));
 //memcpy(Values[index],OldValues[index+1],(nValues-index)*sizeof(SADBObjectStorage *));
 for(int i=0;i<index;i++)Values[i]=OldValues[i];
 for(int i=index+1;i<=nValues;i++)Values[i-1]=OldValues[i];
 free(OldValues);
 pthread_mutex_unlock(&DataAllocationMutex);
 RemovedValue->Free();

 return 0;
}



/// Set Human Readable name of the object
/// @param name
///   Name
/// @return 0 on success
int SADBObject::SetName(const char * name){
 pthread_mutex_lock(&InfoMutex);
 int Length=strlen(name);
 if(Name!=NULL)free(Name);
 Name=(u8*)malloc(Length+1);//+1 to include '\0'
 if(Name==NULL)return 1;
 memcpy(Name,name,Length+1);
 pthread_mutex_unlock(&InfoMutex);
 return 0;
}

/// Set Human Readable description of the object
/// @param desc
///   Description
/// @return 0 on success
int SADBObject::SetDescription(const char * desc){
 pthread_mutex_lock(&InfoMutex);
 int Length=strlen(desc);
 if(Description!=NULL)free(Description);
 Description=(u8*)malloc(Length+1);//+1 to include '\0'
 if(Description==NULL)return 1;
 memcpy(Description,desc,Length+1);
 pthread_mutex_unlock(&InfoMutex);
 return 0;
}

///Get the index of the value that has the timestamp right before (or equal) to the target
///@note Should rewrite this with a better search algorithm, Currently O(n)
///@param Target Target time
///@return Index of the value, or -1 on error
int SADBObject::FindTimestampIndex(Timestamp*Target){
 int Position=nValues-1;
 pthread_mutex_lock(&DataAllocationMutex);//Make sure that the new values are not being allocated
  for(int i=0;i<nValues;i++){
    SADBObjectStorage*Value=Values[i];
    pthread_mutex_lock(&(Value->Mutex));
    if(TimestampDifference(Target,&(Value->Time))>0){
     pthread_mutex_unlock(&(Value->Mutex));//make sure to unlock before breaking
     Position=i;
     break;
    }
    pthread_mutex_unlock(&(Value->Mutex));
  }
 pthread_mutex_unlock(&DataAllocationMutex);
 return Position;
}

///Get the Latest timestanp of the object
///@param T Pointer to a timestamp object
void SADBObject::LatestTimestamp(Timestamp*T){
 SADBObjectStorage*Value=NULL;
 pthread_mutex_lock(&DataAllocationMutex);//Make sure that the new value is not being allocated
 if(nValues>0)Value=Values[nValues-1];//a value exists
 pthread_mutex_unlock(&DataAllocationMutex);
 if(Value!=NULL){
  pthread_mutex_lock(&(Value->Mutex));
   T->Second=(Value->Time).Second;
   T->Millisecond=(Value->Time).Millisecond;
  pthread_mutex_unlock(&(Value->Mutex));
 }else{
   T->Second=0;
   T->Millisecond=0;
 }
}

///Get the timestanp following the target timestamp
///@param Target Pointer to a timestamp object (target)
///@param T Pointer to a timestamp object (return)
void SADBObject::NextTimestamp(Timestamp*Target,Timestamp*T){
 T->Second=0;T->Millisecond=0;//Default initialization
 int ValueIndex=FindTimestampIndex(Target);
 if(ValueIndex==-1)return;//No values
 if(ValueIndex==nValues-1){//Last value, return the latest timestamp
  SADBObjectStorage*Value=Values[nValues-1];
  pthread_mutex_lock(&(Value->Mutex));
   T->Second=(Value->Time).Second;
   T->Millisecond=(Value->Time).Millisecond;
  pthread_mutex_unlock(&(Value->Mutex));
 }else{//Normal Operation
  SADBObjectStorage*Value=Values[ValueIndex+1];
  pthread_mutex_lock(&(Value->Mutex));
   T->Second=(Value->Time).Second;
   T->Millisecond=(Value->Time).Millisecond;
  pthread_mutex_unlock(&(Value->Mutex));
 }
}

///Get the timestanp before the target timestamp
///@param Target Pointer to a timestamp object (target)
///@param T Pointer to a timestamp object (return)
void SADBObject::PreviousTimestamp(Timestamp*Target,Timestamp*T){
 T->Second=0;T->Millisecond=0;//Default initialization
 int ValueIndex=FindTimestampIndex(Target);
 if(ValueIndex==-1)return;//No values
 if(ValueIndex==0){//First value, return the first timestamp
  SADBObjectStorage*Value=Values[0];
  pthread_mutex_lock(&(Value->Mutex));
   T->Second=(Value->Time).Second;
   T->Millisecond=(Value->Time).Millisecond;
  pthread_mutex_unlock(&(Value->Mutex));
 }else{//Normal Operation
  SADBObjectStorage*Value=Values[ValueIndex-1];
  pthread_mutex_lock(&(Value->Mutex));
   T->Second=(Value->Time).Second;
   T->Millisecond=(Value->Time).Millisecond;
  pthread_mutex_unlock(&(Value->Mutex));
 }
}