#include "SADB.hpp"

/// Constructor, initializes object
/// @see SADBObjectStorage::Initialize
SADBObjectStorage::SADBObjectStorage(void){Initialize();}

/// Destructor, deinitializes object
/// @see SADBObjectStorage::Free
SADBObjectStorage::~SADBObjectStorage(void){Free();}

/// Initializes object
void SADBObjectStorage::Initialize(void){
 Data=NULL;Dimensions=NULL;
 pthread_mutex_init(&Mutex,NULL);
}

/// Frees allocated memory
void SADBObjectStorage::Free(void){
 if(Data!=NULL)free(Data);Data=NULL;
 if(Dimensions!=NULL)free(Dimensions);Dimensions=NULL;
 pthread_mutex_destroy(&Mutex);
}

/// Sets the type of data that the object will contain
/// @param Type
///   Data Type
void SADBObjectStorage::SetDataType(enum BODataType Type){
 pthread_mutex_lock(&Mutex);
 DataType=Type;
 switch(DataType){
  case Byte:	ElementSize=1;break;
  case Integer:	ElementSize=4;break;
  case Float:	ElementSize=4;break;
  case Double:	ElementSize=8;break;
  default:	ElementSize=1;//Default to single byte
 }
 pthread_mutex_unlock(&Mutex);
}

/// Sets the size of the object by defining its dimensionality and its dimensions
/// @param D
///   Number of Dimensions
/// @param ...
///   Dimensions (Height, Width, etc)
/// @return 0 on success
int SADBObjectStorage::SetDimensionsf(int D,...){
 pthread_mutex_lock(&Mutex);
 Dimensionality=D;
 Dimensions=(u32 *)realloc(Dimensions,D*sizeof(int));//Allocate array of dimensions
 if(Dimensions==NULL){pthread_mutex_unlock(&Mutex);return 1;}
 va_list DimensionsList;va_start(DimensionsList,D);//Setup variadic argument list
 TotalNumberOfElements=1;
 for(int i=0;i<D;i++){//Set Dimensions
  Dimensions[i]=va_arg(DimensionsList,int);
  TotalNumberOfElements*=Dimensions[i];
 }
 Data=realloc(Data,TotalNumberOfElements*ElementSize);//Allocate Memory Space to hold the object data
 if(Data==NULL){pthread_mutex_unlock(&Mutex);return 2;}
 pthread_mutex_unlock(&Mutex);
 return 0;
}

/// Sets the size of the object by defining its dimensionality and its dimensions
/// @param D
///   Number of Dimensions
/// @param Ds
///   List of Dimensions (Height, Width, etc)
/// @return 0 on success
int SADBObjectStorage::SetDimensions(int D,u32*Ds){
 pthread_mutex_lock(&Mutex);
 Dimensionality=D;
 Dimensions=(u32 *)realloc(Dimensions,D*sizeof(u32));//Allocate array of dimensions
 if(Dimensions==NULL){pthread_mutex_unlock(&Mutex);return 1;}
 TotalNumberOfElements=1;
 for(int i=0;i<D;i++){//Set Dimensions
  Dimensions[i]=Ds[i];
  TotalNumberOfElements*=Dimensions[i];
 }
 Data=realloc(Data,TotalNumberOfElements*ElementSize);//Allocate Memory Space to hold the object data
 if(Data==NULL){pthread_mutex_unlock(&Mutex);return 2;}
 pthread_mutex_unlock(&Mutex);
 return 0;
}

/// Sets the value of an element in the object
/// @param Value
///   The final value of the element
/// @param ...
///   Position of the element
void SADBObjectStorage::SetElement(void * Value,...){
 pthread_mutex_lock(&Mutex);
 va_list Position;va_start(Position,Value);
 Pointer p=(Pointer)Data;//Pointer to the element
 //Get Offset of the element in the buffer in the buffer
 for(int i=0;i<Dimensionality-1;i++){
  u32 Pos=va_arg(Position,int);
  if(Pos>Dimensions[i]){pthread_mutex_unlock(&Mutex);return;}
  p+=Pos*Dimensions[i]*ElementSize;
 }
 p+=va_arg(Position,int)*ElementSize;//Offset in the last dimension
 switch(DataType){//Convert data at given memory location to proper type
  case Byte:	*((u8*)p)=*((u8*)Value);break;
  case Integer:	*((u32*)p)=*((u32*)Value);break;
  case Float:	*((fl*)p)=*((fl*)Value);break;
  case Double:	*((dfl*)p)=*((dfl*)Value);break;
  default:	*((u8*)p)=*((u8*)Value);//Default to single byte
 }
 pthread_mutex_unlock(&Mutex);
}

/// Gets the value of an element in the object
/// @param Value
///   Pointer to a memory location to store the requested value
/// @param ...
///   Position of the element
void SADBObjectStorage::GetElement(void * ReturnValue,...){
 pthread_mutex_lock(&Mutex);
 va_list Position;va_start(Position,ReturnValue);
 Pointer p=(Pointer)Data;//Pointer to the element
 //Get Offset of the element in the buffer in the buffer
 for(int i=0;i<Dimensionality-1;i++){
  u32 Pos=va_arg(Position,int);
  if(Pos>Dimensions[i]){pthread_mutex_unlock(&Mutex);return;}
  p+=Pos*Dimensions[i]*ElementSize;
 }
 p+=va_arg(Position,int)*ElementSize;//Offset in the last dimension
 switch(DataType){//Convert data at given memory location to proper type
  case Byte:	*((u8*)ReturnValue)=*((u8*)p);break;
  case Integer:	*((u32*)ReturnValue)=*((u32*)p);break;
  case Float:	*((fl*)ReturnValue)=*((fl*)p);break;
  case Double:	*((dfl*)ReturnValue)=*((dfl*)p);break;
  default:	*((u8*)ReturnValue)=*((u8*)p);//Default to single byte
 }
 pthread_mutex_unlock(&Mutex);
}

/// Gets the value of an element in the object as a double
/// @param Index
///   Position of the element
/// @return Value requested as a double
double SADBObjectStorage::GetDoubleElement(int Index){
 double R;
 pthread_mutex_lock(&Mutex);
 Pointer p=(Pointer)Data+Index*ElementSize;;//Pointer to the element
 switch(DataType){//Convert data at given memory location to proper type
  case Byte:	R=(double)*((u8*)p);break;
  case Integer:	R=(double)*((u32*)p);break;
  case Float:	R=(double)*((fl*)p);break;
  case Double:	R=(double)*((dfl*)p);break;
  default:	R=(double)*((u8*)p);//Default to single byte
 }
 pthread_mutex_unlock(&Mutex);
 return R;
}

/// Sets the value of an element in the object
/// @param Value
///   Value to set
/// @param Index
///   Position of the element
void SADBObjectStorage::SetDoubleElement(double Value,int Index){
 pthread_mutex_lock(&Mutex);
 Pointer p=(Pointer)Data+Index*ElementSize;;//Pointer to the element
 switch(DataType){//Convert data at given memory location to proper type
  case Byte:	*((u8*)p)=(u8)Value;break;
  case Integer:	*((u32*)p)=(u32)Value;break;
  case Float:	*((fl*)p)=(fl)Value;break;
  case Double:	*((dfl*)p)=(dfl)Value;break;
  default:	*((u8*)p)=(u8)Value;//Default to single byte
 }
 pthread_mutex_unlock(&Mutex);
}


/// Set the timestamp
/// @param time
///  Timestamp
void SADBObjectStorage::SetTimestamp(Timestamp time){
 pthread_mutex_lock(&Mutex);
 Time.Second=time.Second;
 Time.Millisecond=time.Millisecond;
 pthread_mutex_unlock(&Mutex);
}

/// Set the timestamp
/// @param Second
///  Seconds
/// @param Millisecond
///  Milliseconds
void SADBObjectStorage::SetTimestamp(u64 Second,u16 Millisecond){
 pthread_mutex_lock(&Mutex);
 Time.Second=Second;
 Time.Millisecond=Millisecond;
 pthread_mutex_unlock(&Mutex);
}


/// Print Human Readable information about the object
void SADBObjectStorage::Information(void){
 printf("Contains %d dimensions\n",Dimensionality);
 printf("Dimensions: ");
 for(int i=0;i<Dimensionality;i++)printf("%d ",Dimensions[i]);
 printf("\n");

 switch(DataType){//Convert data at given memory location to proper type
  case Byte:	printf("Data Type: Byte\n");break;
  case Integer:	printf("Data Type: Integer\n");break;
  case Float:	printf("Data Type: Float\n");break;
  case Double:	printf("Data Type: Double\n");break;
  default:	printf("Data Type: Defaulted to Byte\n");
 }

 if(TotalNumberOfElements<100){
  printf("Raw Data: \n");
  Pointer p=(Pointer)Data;
  for(int i=0;i<TotalNumberOfElements;i++){
   switch(DataType){//Convert data at given memory location to proper type
    case Byte:		printf("%03d\t",*((u8*)p));	p+=1;break;
    case Integer:	printf("%d\t",	*((u32*)p));	p+=4;break;
    case Float:		printf("%f\t",	*((fl*)p));	p+=4;break;
    case Double:	printf("%lf\t",	*((dfl*)p));	p+=8;break;
    default:		printf("%03d\t",*((u8*)p));	p+=1;
   }
  }  
  printf("\n");
 }

}
