

#ifndef __SADBServer__
#define __SADBServer__

///Holds information about a single SADB client
typedef struct{
 int Socket;///<Socket FD
 SADBServer*Board;///<Pointer to the parent board
}SADBClientInfo;


///Main SADB server class
class SADBServer {
 public:
  SADBServer(void);
  ~SADBServer(void);

  void Initialize(void);///<Initializes object, called by constructor
  void Free(void);///<Frees allocated memory, called by destructor
  void Reset(void);///<Soft-Reset the SADB server, some memory might not get deallocated probably, should not be used too often

  void Start(int port=DefaultSADBPort);///<Start the server
  void Stop(void);///<Stop the server
  
  
  
  



  //Object management
  ///Determine if an OUID is valid or not
  ///@param OUID OUID to test
  ///@return 0 if not valid
  int IsValidOUID(u64 OUID);
  ///Create a new blank object
  ///@return ID of the newly created object
  u64 CreateObject(void);
  ///Removes an object from the list.  Does not actually remove the object because then we'd need to dealloc/realloc, but sets the active bit to 0 and deallocates object memory (data buffers).  We could add the dealloc/realloc like in the client thread removal, but it would be slow if there's a lot of objects 
  ///@param ID ID of the object to be removed
  ///@return 0 on success
  int RemoveObject(u64 ID);
  
  //Category Management
  ///Determine if an CUID is valid or not
  ///@param CUID CUID to test
  ///@return 0 if not valid
  int IsValidCUID(u64 CUID);
  ///Create a new blank category
  ///@return ID of the newly created object
  u64 CreateCategory(void);
  ///Removes a category from the list.  Does not actually remove the category because then we'd need to dealloc/realloc, which would be slow if there's a lot of categories. 
  ///@param ID ID of the category to be removed
  ///@return 0 on success
  int RemoveCategory(u64 ID);
  ///Get the CUID of a category from its name
  ///@param Name Name of the category
  ///@return CUID
  u64 GetCategoryByName(u8*Name);




  
  //Server functions
  ///Answer to the ping command
  ///@param Info Client Information
  void CMDPing(SADBClientInfo*Info);
  ///Soft Reset the SADB Server
  ///@param Info Client Information
  void CMDReset(SADBClientInfo*Info);
  ///Find object(s) based on a query
  ///@param Info Client Information
  ///@param String String containing the query
  void CMDFindObjects(SADBClientInfo*Info,u8*String);
  ///Create a new object on the blackboard
  ///@param Info Client Information
  void CMDCreateObject(SADBClientInfo*Info);
  ///Remove an object from the blackboard
  ///@param Info Client Information
  ///@param OUID ID of the object to be removed
  void CMDDeleteObject(SADBClientInfo*Info,u64 OUID);
  ///Check if an object is valid (exists & not deleted)
  ///@param Info Client Information
  ///@param OUID ID of the object
  void CMDIsValidObject(SADBClientInfo*Info,u64 OUID);
  ///Set the name of an object on the blackboard
  ///@param Info Client Information
  ///@param OUID ID of the object to edit
  ///@param Name Null-terminated string to set
  void CMDSetObjectName(SADBClientInfo*Info,u64 OUID,u8 * Name);
  ///Get the name of an object on the blackboard
  ///@param Info Client Information
  ///@param OUID ID of the object to edit
  void CMDGetObjectName(SADBClientInfo*Info,u64 OUID);
  ///Set the description of an object on the blackboard
  ///@param Info Client Information
  ///@param OUID ID of the object to edit
  ///@param Description Null-terminated string to set
  void CMDSetObjectDescription(SADBClientInfo*Info,u64 OUID,u8 * Description);
  ///Get the description of an object on the blackboard
  ///@param Info Client Information
  ///@param OUID ID of the object to edit
  void CMDGetObjectDescription(SADBClientInfo*Info,u64 OUID);
  ///Get the OUID of an object from its name
  ///@param Info Client Information
  ///@param Name Name of the object
  void CMDGetObjectByName(SADBClientInfo*Info,u8*Name);
  ///The client sends a value of the object
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Datatype Type of the data
  ///@param Second Second part of the timestamp
  ///@param Millisecond Millisecond part of the timestamp
  ///@param Data Data Received
  void CMDSendObject(SADBClientInfo*Info,u64 OUID,u32 Datatype,u64 Second,u16 Millisecond,u8*Data);
  ///Send the value of an object
  ///@param Info Client Information
  ///@param OUID ID of the object
  void CMDGetLatestObjectValue(SADBClientInfo*Info,u64 OUID);
  ///Send the latest timestamp of an object
  ///@param Info Client Information
  ///@param OUID ID of the object
  void CMDGetLatestObjectTimestamp(SADBClientInfo*Info,u64 OUID);
  ///Send the next timestamp of an object, relative to the supplied target timestamp
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target "Current" timestamp
  void CMDGetNextObjectTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target);
  ///Send the previous timestamp of an object, relative to the supplied target timestamp
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target "Current" timestamp
  void CMDGetPreviousObjectTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target);
  ///Send the closest timestamp to a target, relative to the supplied target timestamp
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target "Current" timestamp
  void CMDGetNearestObjectTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target);
  ///Send the value of an object at the nearest timestamp to the target
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target Target Timestamp
  void CMDGetObjectValueAtNearestTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target);
  ///Remove The Earliest Value of an object
  ///@param Info Client Information
  ///@param OUID ID of the object
  void CMDRemoveOldestValue(SADBClientInfo*Info,u64 OUID);
  ///Keep only a certain amount of Values
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param nValues Number of values to keep
  void CMDKeepOnlyLatestValues(SADBClientInfo*Info,u64 OUID,int nValues);
  ///Send the value of an object at the requested timestamp by interpolation
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target Target Timestamp
  void CMDGetObjectValueAtTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target,u64 Argument);
  ///Send the value of an object at the timestamp following the requested timestamp
  ///@note If there is no values after the requested timestamp, the latest value is returned
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target Target Timestamp
  void CMDGetObjectValueAtNextTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target);
  ///Send the value of an object at the timestamp following the requested timestamp
  ///@note If there is no values before the requested timestamp, the first value is returned
  ///@param Info Client Information
  ///@param OUID ID of the object
  ///@param Target Target Timestamp
  void CMDGetObjectValueAtPreviousTimestamp(SADBClientInfo*Info,u64 OUID,Timestamp Target);
  ///Sends the ID of the next object that will be added to the board
  ///@param Info Client Information
  void CMDGetNextObjectID(SADBClientInfo*Info);
  ///Create a new category on the blackboard
  ///@param Info Client Information
  void CMDCreateCategory(SADBClientInfo*Info);
  ///Remove a category from the blackboard
  ///@param Info Client Information
  ///@param CUID ID of the category to be removed
  void CMDDeleteCategory(SADBClientInfo*Info,u64 CUID);
  ///Check if a category is valid (exists & not deleted)
  ///@param Info Client Information
  ///@param CUID ID of the category
  void CMDIsValidCategory(SADBClientInfo*Info,u64 CUID);
  ///Set the name of a category on the blackboard
  ///@param Info Client Information
  ///@param CUID ID of the category to edit
  ///@param Name Null-terminated string to set
  void CMDSetCategoryName(SADBClientInfo*Info,u64 CUID,u8 * Name);
  ///Get the name of a category on the blackboard
  ///@param Info Client Information
  ///@param CUID ID of the category to edit
  void CMDGetCategoryName(SADBClientInfo*Info,u64 CUID);
  ///Set the description of a category on the blackboard
  ///@param Info Client Information
  ///@param CUID ID of the category to edit
  ///@param Name Null-terminated string to set
  void CMDSetCategoryDescription(SADBClientInfo*Info,u64 CUID,u8 * Description);
  ///Get the description of a category on the blackboard
  ///@param Info Client Information
  ///@param CUID ID of the category to edit
  void CMDGetCategoryDescription(SADBClientInfo*Info,u64 CUID);
  ///Get the CUID of a category from its name
  ///@param Info Client Information
  ///@param Name Name of the category
  void CMDGetCategoryByName(SADBClientInfo*Info,u8*Name);
  ///Add an object to a category
  ///@param Info Client Information
  ///@param CUID ID of the category
  ///@param OUID ID of the object
  void CMDAddObjectToCategory(SADBClientInfo*Info,u64 CUID,u64 OUID);
  ///Remove an object from a category
  ///@param Info Client Information
  ///@param CUID ID of the category
  ///@param OUID ID of the object
  void CMDRemoveObjectFromCategory(SADBClientInfo*Info,u64 CUID,u64 OUID);
  /// Sends the ID of the next category that will be added to the board
  ///@param Info Client Information
  void CMDGetNextCategoryID(SADBClientInfo*Info);

  //Object related
  SADBObject**Objects;///<List of objects currently residing on the blackboard
  u64 nObjects;///<Number of objects (length of the Objects list)
  u8*ObjectStates;///<Object State Boolean List (active vs deleted objects)
  u64 CurrentObjectID;///<Next available unique ID, increases sequentially
  pthread_mutex_t ObjectAllocationMutex;///<used to prevent two threads to (de)allocate objects at the same time
  
  //Category related
  SADBCategory**Categories;///<List of categories currently residing on the blackboard
  u64 nCategories;///<Number of categories (length of the categories list)
  u8*CategoryStates;///<Category State List (active vs deleted objects)
  u64 CurrentCategoryID;///<Next available unique ID, increases sequentially
  pthread_mutex_t CategoryAllocationMutex;///<used to prevent two threads to (de)allocate categories
  
  int Port;///<Port Number
  
 private:
  int StopRequest;///<Boolean Stop Request
  Network Server;///<Handles most Network Related Stuff
  pthread_t ServerMainThreadID;///<pthead object for the main thread
  ///Static function to start the main thread
  ///@param context void casted pointer to a SADBServer object
  static void *StartMainThreadProxy(void *context);
  void Main(void);///<Main Method for the server thread
  ///Client Handling function
  ///@param ClientInfo void casted pointer to a HTTPClientInfo object
  static void *SADBClientHandler(void*ClientInfo);
};
#endif 

